""" Application.py

This is the main module that binds everything together.

"""
__copyright__ = "Copyright (c) 2002-2005 Free Software Foundation, Inc."
__license__ = """ GNU General Public License

This program is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA. """

import time
from xml.sax import saxutils
import pygtk
pygtk.require('2.0')
import gobject
import gtk
import gtk.glade
import gnome
import gconf
import utils
import subscribe
import Event
import MessageManager
import FeedList
import FeedCategoryList
import Config
import error
from ItemView import ItemView
from ItemList import ItemListPresenter, ItemListView
from FeedListView import FeedsView, FeedsPresenter
from OfflineToggle import OfflineToggle
from CategorySelector import CategoryPresenter, CategoryView
from MainloopManager import MainloopManager
from FeedPropertiesDialog import FeedPropertiesDialog
from PreferencesDialog import PreferencesDialog
from Find import FindPresenter, FindView
import OPMLImport
import MVP
import ItemStore
import ImageCache
import PollManager
import dialogs
import constants

class StatusPresenter(MVP.BasicPresenter):
    def _initialize(self):
        self._mmgr = MessageManager.get_instance()
        self._mmgr.signal_connect(Event.StatusDisplaySignal,
                                  self._display)
        return

    def _display(self, signal):
        cid =  self._view.get_context_id("straw_main")
        self._view.pop(cid)
        self._view.push(cid, self._mmgr.read_message())
        return

class ErrorPresenter:
    def __init__(self, widget):
        self._widget = widget
        self._tooltips = gtk.Tooltips()
        self._text = ''
        self._curr_feed = None
        self._curr_category = None
        fclist = FeedCategoryList.get_instance()
        fclist.signal_connect(Event.FeedCategoryChangedSignal,
                              self._category_changed)

    def display_feed_error(self, feed):
        if self._curr_feed is not None:
            self._curr_feed.signal_disconnect(Event.FeedErrorStatusChangedSignal,
                                              self._error_status_changed)
        self._curr_feed = feed
        self._curr_feed.signal_connect(Event.FeedErrorStatusChangedSignal,
                                       self._error_status_changed)
        self._update_view()
        return

    def display_category_error(self, category):
        if category:
            self._curr_category = category
            self._update_view()
        return

    def _update_view(self):
        text = list()
        if self._curr_category and self._curr_category.subscription and self._curr_category.subscription.error:
            text.append(_("Category error:"))
            text.append(self._curr_category.subscription.error)
        if self._curr_feed and self._curr_feed.error:
            text.append(_("Feed Error:"))
            text.append(self._curr_feed.error)
        if text:
            t = "\n".join(text)
            self._tooltips.set_tip(self._widget,t,t)
            self._tooltips.enable()
            self._widget.show()
        else:
            self._tooltips.disable()
            self._widget.hide()
        return

    def _category_changed(self, signal):
        if signal.sender is self._curr_category:
            self.display_category_error(self._curr_category)

    def _error_status_changed(self, signal):
        if signal.sender is self._curr_feed:
            self.display_feed_error(signal.sender)
        return

    def hide(self):
        self._widget.hide()

class FeedInfoView(MVP.WidgetView):
    def _initialize(self):
        widget_tree = gtk.glade.get_widget_tree(self._widget)
        self._title = widget_tree.get_widget('feed_item_label')
        self._refresh = widget_tree.get_widget('feed_refresh_label')

    def set_title(self, title):
        self._title.set_label(title)
        return

    def set_refresh(self, refresh):
        self._refresh.set_label(refresh)
        self._refresh.show()
        return

    def hide(self):
        self._widget.hide()
        return

    def show(self):
        self._widget.show()
        return

class FeedInfoPresenter(MVP.BasicPresenter):
    def display(self, feed):
        next = feed.next_refresh
        if feed:
            text = _("Next Refresh: %s") % utils.format_date(time.gmtime(next))
            self._view.set_refresh(text)
            self._view.set_title(_("%s") % saxutils.escape(feed.title))
            self._view.show()
        else:
            self._view.hide()
        return

    def hide(self):
        self._view.set_title("")
        self._view.set_refresh("")

    def show(self): self._view.show()

class MenuFeedPropsPresenter(MVP.BasicPresenter):
    def set_sensitive(self, s):
        self._view.set_sensitive(s)
        return

class ToolbarView(MVP.WidgetView):
    """ Widget: gtk.Toolbar"""
    GCONF_DESKTOP_INTERFACE = "/desktop/gnome/interface"

    def _initialize(self):
        client = gconf.client_get_default()
        if not client.dir_exists(self.GCONF_DESKTOP_INTERFACE):
            return
        client.add_dir(self.GCONF_DESKTOP_INTERFACE,
                       gconf.CLIENT_PRELOAD_NONE)
        client.notify_add(self.GCONF_DESKTOP_INTERFACE+"/toolbar_style",
                          self._toolbar_style_changed)
        self._widget.set_tooltips(True)

    def _toolbar_style_changed(self, client, notify_id, entry, *args):
        value = entry.value.get_string()
        self._presenter.style_changed(value)

    def set_style(self, style):
        self._widget.set_style(style)

    def get_style(self):
        client = gconf.client_get_default()
        current_style = client.get_string(self.GCONF_DESKTOP_INTERFACE+"/toolbar_style")
        return current_style

class ToolbarPresenter(MVP.BasicPresenter):
    STYLES = {'both':gtk.TOOLBAR_BOTH,
              'text':gtk.TOOLBAR_TEXT,
              'icons':gtk.TOOLBAR_ICONS,
              'both-horiz':gtk.TOOLBAR_BOTH_HORIZ}

    def _initialize(self):
        style = self._view.get_style()
        if style in self.STYLES.keys():
            self._view.set_style(self.STYLES[style])
        return

    def style_changed(self, value):
        if value in self.STYLES.keys():
            self._view.set_style(self.STYLES[value])
        return

class ApplicationPresenter(MVP.BasicPresenter):
    def _initialize(self):
        self._curr_category = None
        self._curr_feed = None
        self._curr_item = None
        self._init_widgets()
        self._init_presenters()
        self._init_signals()
        self._prefs_dialog = None
        self._view.present()

    def _init_widgets(self):
        widget_tree = self._view.get_widget_tree()
        self._itemlist_view_notebook = widget_tree.get_widget('mode_view_notebook')
        self._feedlist_view_notebook = widget_tree.get_widget('left_mode_view_notebook')

    def _init_presenters(self):
        widget_tree = self._view.get_widget_tree()
        toolbar_presenter = ToolbarPresenter(view=ToolbarView(widget_tree.get_widget('toolbar_default')))
        self._feed_list_presenter = FeedsPresenter(view=FeedsView(widget_tree.get_widget('feed_selection_treeview')))
        self._category_selector = CategoryPresenter(view=CategoryView(widget_tree.get_widget('category_combo')))
        self._error_presenter = ErrorPresenter(
            widget_tree.get_widget('statusbar_error_indicator'))
        self._offline_presenter = OfflineToggle(
            widget_tree.get_widget('offline_toggle'))
        self._itemlist_presenter = ItemListPresenter(view = ItemListView(widget_tree.get_widget('item_selection_treeview')))
        self._item_view = ItemView(widget_tree.get_widget('item_view_container'))
        self._status_presenter = StatusPresenter(view = widget_tree.get_widget("main_statusbar"))
        self._menufp_presenter = MenuFeedPropsPresenter( view = widget_tree.get_widget('menu_feed_properties'))
        self._menufp_presenter.set_sensitive(False)
        self._feedinfo_presenter = FeedInfoPresenter(view = FeedInfoView(widget_tree.get_widget('article_refresh_hbox')))
        self._find_presenter = FindPresenter(view=FindView(widget_tree.get_widget("find_vbox")))
        return

    def _init_signals(self):
        self._feed_list_presenter.signal_connect(Event.FeedSelectionChangedSignal,
                                                 self._feed_selection_changed)
        self._feed_list_presenter.signal_connect(Event.FeedsEmptySignal,
                                                 self._feeds_empty_cb)
        self._itemlist_presenter.signal_connect(Event.ItemSelectionChangedSignal,
                                                self._item_view.item_selection_changed)
        self._category_selector.signal_connect(Event.CategorySelectionChangedSignal,
                                               self._category_selection_changed)
        return

    def _feed_selection_changed(self, signal):
        if signal.old and self._curr_feed:
            self._curr_feed.unload_contents()
        if signal.new:
            if self._curr_feed:
                self._curr_feed.signal_disconnect(Event.FeedPolledSignal,
                                                  self._feed_polled)
                self._curr_feed.signal_disconnect(Event.ItemsAddedSignal,
                                                  self._feed_items_added)
            self._curr_feed = signal.new
            self._curr_feed.signal_connect(Event.FeedPolledSignal,
                                           self._feed_polled)
            self._curr_feed.signal_connect(Event.ItemsAddedSignal,
                                           self._feed_items_added)
            self._display_feed(self._curr_feed)
        return

    def _feed_polled(self, signal):
        self._feedinfo_presenter.display(signal.sender)
        return

    def _feed_items_added(self, signal):
        self._itemlist_presenter.display_feed_items(self._curr_feed, False)

    def _category_selection_changed(self, signal):
        self._display_category_feeds(signal.current)
        return

    def _display_category_feeds(self, category, *args):
        self._curr_category = category
        if category:
            self._feed_list_presenter.display_category_feeds(category)
            self._error_presenter.display_category_error(category)
        else:
            self._feed_list_presenter.display_empty_category()
            self._menufp_presenter.set_sensitive(False)
            self._itemlist_presenter.display_empty_feed()
            self._item_view.display_empty_feed()
        return

    def _display_feed(self, feed, select_first = 1):
        if feed and feed.number_of_items < 1:
            self._item_view.display_empty_feed()
        self._error_presenter.display_feed_error(feed)
        self._feedinfo_presenter.display(feed)
        self._itemlist_presenter.display_feed_items(feed, select_first)
        self._menufp_presenter.set_sensitive(True)
        return

    def _feeds_empty_cb(self, signal):
        self._itemlist_presenter.display_empty_feed()
        self._item_view.display_empty_feed()
        self._feedinfo_presenter.hide()
        self._error_presenter.hide()
        self._menufp_presenter.set_sensitive(False)
        return

    def copy_itemview_text_selection(self):
        utils.set_clipboard_text(self._item_view.get_selected_text())

    def check_allocation(self, widget, event):
        config = Config.get_instance()
        def check_size((width, height, widget)):
            if width == widget.allocation.width and height == widget.allocation.height:
                config.main_window_size = (width, height)
        if event.width != widget.allocation.width or event.height != widget.allocation.height:
            gobject.timeout_add(1000, check_size, (
                (event.width, event.height, widget)))

    def check_main_pane_position(self, widget):
        config = Config.get_instance()
        def check_position((position, widget)):
            if position == widget.get_position():
                config.main_pane_position = position
        pos = widget.get_position()
        if pos != config.main_pane_position:
            gobject.timeout_add(1000, check_position, (pos, widget))

    def check_sub_pane_position(self, widget):
        config = Config.get_instance()
        def check_position((position, widget)):
            if position == widget.get_position():
                config.sub_pane_position = position
        pos = widget.get_position()
        if pos != config.sub_pane_position:
            gobject.timeout_add(1000, check_position, (pos, widget))

    def credits(self):
        return dialogs.credits()

    def poll_all(self):
        if self._warn_if_offline():
            fclist = FeedCategoryList.get_instance()
            self._poll_categories(fclist.all_categories)
        return

    def poll_current_category(self):
        if self._warn_if_offline():
            self._poll_categories([self._curr_category])
        return

    def poll_current_feed(self):
        if self._warn_if_offline():
            pm = PollManager.get_instance()
            pm.poll([self._curr_feed])
        return

    def mark_feed_as_read(self):
        self._curr_feed.mark_all_read()

    def mark_all_as_read(self):
        mmgr = MainloopManager.get_instance()
        flist = FeedList.get_instance().flatten_list()
        mmgr.call_pending()
        for feed in flist:
            feed.mark_all_read()
            if feed is self._curr_feed:
                continue
            feed.unload_contents()
            mmgr.call_pending()

    def remove_selected_feed(self):
        self._feed_list_presenter.remove_selected_feed()

    def show_search(self):
        self._itemlist_presenter.signal_disconnect(Event.ItemSelectionChangedSignal,
                                                   self._item_view.item_selection_changed)
        self._find_presenter.item_list.signal_connect(Event.ItemSelectionChangedSignal,
                                                      self._item_view.item_selection_changed)
        self._item_view.display_empty_search()
        self._itemlist_view_notebook.set_current_page(1)
        self._feedlist_view_notebook.set_current_page(1)
        self._feedinfo_presenter.hide()

    def hide_search(self):
        self._find_presenter.clear()
        self._itemlist_view_notebook.set_current_page(0)
        self._feedlist_view_notebook.set_current_page(0)
        self._feedinfo_presenter.show()
        self._find_presenter.item_list.signal_disconnect(Event.ItemSelectionChangedSignal,
                                                     self._item_view.item_selection_changed)
        self._itemlist_presenter.signal_connect(Event.ItemSelectionChangedSignal,
                                                self._item_view.item_selection_changed)

    def _poll_categories(self, fclist):
        pm = PollManager.get_instance()
        pm.poll_categories(fclist)
        return

    def _warn_if_offline(self):
        config = Config.get_instance()
        will_poll = False
        if config.offline:
            response = self._view.show_offline_dialog()
            if response == gtk.RESPONSE_OK:
                config.offline = not config.offline
                will_poll = True
        else:
            will_poll = True
        return will_poll

    def _get_next_category(self, category = None):
        if not category:
            category = self._curr_category
        fclist = FeedCategoryList.get_instance()
        allcats = fclist.user_categories + list(fclist.pseudo_categories)
        if category is None:
            category = allcats[0]
        else:
            index = allcats.index(category)
            if index < len(allcats) - 1:
                index += 1
            else:
                index = 0
            category = allcats[index]
        return category

    def _get_previous_category(self, category = None):
        if category is None:
            category = self._curr_category
        fclist = FeedCategoryList.get_instance()
        allcats = fclist.user_categories + list(fclist.pseudo_categories)
        if category is None:
            category = allcats[-1]
        else:
            index = allcats.index(category)
            if index > 0:
                index -= 1
            else:
                index = len(allcats) - 1
            category = allcats[index]
        return category

    def display_previous_category(self, category = None):
        """
        Displays the category before the current selected category
        """
        category = self._get_previous_category(category)
        self._category_selector.category_selected(category)
        self._display_category_feeds(category)

    def display_next_category(self, category = None):
        """
        Display the category after the current selected category
        """
        category = self._get_next_category(category)
        self._category_selector.category_selected(category)
        self._display_category_feeds(category)

    def display_previous_feed(self, item = None):
        """
        Displays the feed before the current selected feed
        """
        self._feed_list_presenter.select_previous_feed()

    def display_next_feed(self, item=None):
        """
        Displays the feed after the current selected feed
        """
        is_next = self._feed_list_presenter.select_next_feed()
        if not is_next:
            self._feed_list_presenter.select_first_feed()
        return

    def display_next_unread_feed(self):
        """
        Displays the next feed with an unread item
        """
        self._feed_list_presenter.select_next_unread_feed()

    def display_previous_item(self, item=None):
        """
        Displays the item before the current selected item. If the item is the
        first item, scrolls to the previous feed
        """
        is_prev = self._itemlist_presenter.select_previous_item()
        if not is_prev:
            # TODO HACK - implement select_previous_feed(select_last=True) ...
            # ... to select previous feed's last item
            self._feed_list_presenter.select_previous_feed()
            self._itemlist_presenter.select_last_item()
        return

    def display_next_item(self, item=None):
        """
        Displays the item after the current selected item. If the item is the
        last item, selectes the next feed. If the current feed is the last
        feed in the list, it goes back and selects the first feed
        """
        is_next = self._itemlist_presenter.select_next_item()
        if not is_next:
            is_next_feed = self._feed_list_presenter.select_next_feed()
            if not is_next_feed:
                self._feed_list_presenter.select_first_feed()
        return

    def scroll_or_display_next_unread_item(self, item=None):
        has_unread_item = False
        if not self._item_view.scroll_down():
            has_unread_item = self._itemlist_presenter.select_next_unread_item()
            if not has_unread_item:
                self._feed_list_presenter.select_next_unread_feed()
        return

    def show_preferences_dialog(self, parent):
        if not self._prefs_dialog:
            xf = utils.find_glade_file()
            xml = gtk.glade.XML(xf, "preferences_dialog", gettext.textdomain())
            self._prefs_dialog = PreferencesDialog(xml, parent)
        self._prefs_dialog.show()

    def show_feed_properties(self, parent):
        fpd = FeedPropertiesDialog.show_feed_properties(parent, self._curr_feed)
        return

    def quit(self):
        gtk.main_quit()
        ItemStore.get_instance().stop()
        return

class ApplicationView(MVP.WidgetView):
    """
    Widget: straw_main
    """
    def _initialize(self):
        self._config = Config.get_instance()
        self._initialize_dnd()
        self._initialize_window_updater()
        self._create_unmodified_accelerator_group()
        self._attach_unmodified_accelerator_group()
        self._initialize_window()
        self._find_toggled = False

    def _initialize_window(self):
        widget_tree = gtk.glade.get_widget_tree(self._widget)
        if self._config.window_maximized:
            self._widget.maximize()
        else:
            # we use resize here since configure-event seems to
            # overwrite the default size if we use set_default_size.
            self._widget.resize(*self._config.main_window_size)
        mmp = widget_tree.get_widget('main_main_pane')
        msp = widget_tree.get_widget('main_sub_pane')
        mmp.set_position(self._config.main_pane_position)
        msp.set_position(self._config.sub_pane_position)

    def _initialize_dnd(self):
        self._widget.drag_dest_set(
                gtk.DEST_DEFAULT_ALL,
                [('_NETSCAPE_URL', 0, 0), ('text/uri-list ', 0, 1),
                 ('x-url/http', 0, 2)],
                gtk.gdk.ACTION_COPY | gtk.gdk.ACTION_MOVE)
        return

    def _initialize_window_updater(self):
        feedlist = FeedList.get_instance()
        feedlist.signal_connect(Event.AllItemsReadSignal,
                                lambda signal: self._update_title(feedlist))
        feedlist.signal_connect(Event.ItemReadSignal,
                                lambda signal: self._update_title(feedlist))
        feedlist.signal_connect(Event.ItemsAddedSignal,
                                lambda signal: self._update_title(feedlist))
        feedlist.signal_connect(Event.FeedsChangedSignal,
                                lambda signal: self._update_title(feedlist))
        return

    def _update_title(self, flist):
        uritems = urfeeds = 0
        sfeeds = "feeds"
        listfeeds = flist.flatten_list()
        for ur in [f.n_items_unread for f in listfeeds]:
            if ur:
                uritems += ur
                urfeeds += 1
        else:
            urfeeds = len(listfeeds)
        if urfeeds < 2:
            sfeeds = "feed"
        item_feed_map = {'uritems': uritems,
                         'urfeeds': urfeeds,
                         'fstring' : sfeeds}
        title = _('%(uritems)d unread in %(urfeeds)d %(fstring)s') % item_feed_map
        self._widget.set_title( title + " - %s" % constants.APPNAME)
        return

    # We have a separate accelerator group for the unmodified and
    # shifted accelerators, that is, stuff like space, N, P, etc. This
    # is so that we can have the find pane work correctly
    def _create_unmodified_accelerator_group(self):
        xml = gtk.glade.get_widget_tree(self._widget)
        agroup = gtk.AccelGroup()
        accels = (('menu_mark_feed_as_read', 'R', gtk.gdk.SHIFT_MASK),
                  ('menu_mark_all_as_read', 'A', gtk.gdk.SHIFT_MASK),
                  ('menu_next', 'N', gtk.gdk.SHIFT_MASK),
                  ('menu_next_unread', ' ', 0),
                  ('menu_previous', 'P', gtk.gdk.SHIFT_MASK))
        for widget_name, key, mask in accels:
            widget = xml.get_widget(widget_name)
            widget.add_accelerator("activate", agroup, ord(key), mask,
                                   gtk.ACCEL_VISIBLE)
        self._unmodified_accelerator_group = agroup

    def _attach_unmodified_accelerator_group(self):
        self._widget.add_accel_group(self._unmodified_accelerator_group)

    def _detach_unmodified_accelerator_group(self):
        self._widget.remove_accel_group(self._unmodified_accelerator_group)

    def _on_straw_main_destroy_event(self, *args):
        return self._presenter.quit()

    def _on_straw_main_delete_event(self, *args):
        return self._presenter.quit()

    def _on_menu_quit_activate(self, *args):
        return self._presenter.quit()

    def _on_straw_main_configure_event(self, widget, event, *args):
        if widget.window.get_state() is not gtk.gdk.WINDOW_STATE_MAXIMIZED:
            self._config.window_maximized = False
            self._presenter.check_allocation(widget, event)
        else:
            self._config.window_maximized = True
        return

    def _on_main_main_pane_size_allocate(self, widget, *args):
        self._presenter.check_main_pane_position(widget)

    def _on_main_sub_pane_size_allocate(self, widget, *args):
        self._presenter.check_sub_pane_position(widget)

    def _on_menu_report_problem_activate(self, menuitem, *args):
        utils.url_show("http://bugzilla.gnome.org/simple-bug-guide.cgi?product=straw")

    def _on_menu_about_activate(self, menuitem, *args):
        widget = self._presenter.credits()
        widget.show()

    def _on_menu_refresh_all_activate(self, *args):
        self._presenter.poll_all()

    def _on_menu_refresh_category_activate(self, *args):
        self._presenter.poll_current_category()

    def _on_menu_refresh_selected_activate(self, *args):
        self._presenter.poll_current_feed()

    def _on_toolbar_refresh_all_button_clicked(self, *args):
        self._presenter.poll_all()

    def _on_menu_add_activate(self, *args):
        subscribe.show(self._widget)

    def _on_toolbar_subscribe_button_clicked(self, *args):
        subscribe.show(self._widget)

    def _on_menu_import_subscriptions_activate(self, *args):
        dialogs.import_subscriptions(self._widget)

    def _on_menu_export_subscriptions_activate(self, *args):
        dialogs.export_subscriptions(self._widget)

    def _on_menu_copy_text_activate(self, *args):
        self._presenter.copy_itemview_text_selection()

    def _on_menu_mark_feed_as_read_activate(self, *args):
        self._presenter.mark_feed_as_read()

    def _on_menu_mark_all_as_read_activate(self, *args):
        self._presenter.mark_all_as_read()

    def _on_find_activate(self, widget, *args):
        xml = gtk.glade.get_widget_tree(self._widget)
        menu_find = xml.get_widget('menu_find')
        accel_label = menu_find.get_child()
        
        if not self._find_toggled:
            self._presenter.show_search()
            self._detach_unmodified_accelerator_group()
            self._find_toggled = True
            
            # save the "Find..." stock text for later recovery
            self._old_label_text = accel_label.get_text()
            accel_label.set_text(_('Return to feed list...'))
            
        else:
            self._presenter.hide_search()
            self._attach_unmodified_accelerator_group()
            self._find_toggled = False
            accel_label.set_text(self._old_label_text)            

    def _on_menu_next_activate(self, *args):
        self._presenter.display_next_item()

    def _on_toolbar_scroll_or_next_button_clicked(self, *args):
        self._presenter.scroll_or_display_next_unread_item()

    def _on_menu_scroll_next_activate(self, *args):
        self._presenter.scroll_or_display_next_unread_item()

    def _on_menu_previous_activate(self, *args):
        self._presenter.display_previous_item()

    def _on_menu_next_feed_unread_activate(self, *args):
        self._presenter.display_next_unread_feed()

    def _on_menu_next_feed_activate(self, *args):
        self._presenter.display_next_feed()

    def _on_menu_previous_feed_activate(self, *args):
        self._presenter.display_previous_feed()

    def _on_menu_remove_selected_feed_activate(self, *args):
        self._presenter.remove_selected_feed()

    def _on_menu_next_category_activate(self, *args):
        self._presenter.display_next_category()

    def _on_menu_previous_category_activate(self, *args):
        self._presenter.display_previous_category()

    def _on_menu_next_category_activate(self, *args):
        self._presenter.display_next_category()

    def _on_menu_previous_category_activate(self, *args):
        self._presenter.display_previous_category()

    def _on_menu_preferences_activate(self, *args):
        self._presenter.show_preferences_dialog(self._widget)

    def _on_menu_feed_properties_activate(self, *args):
        self._presenter.show_feed_properties(self._widget)

    def _on_straw_main_drag_data_received(self, w, context,
                                          x, y, data, info, time):
        if data and data.format == 8:
            url = data.data.split("\n")[0]
            subscribe.show(self._widget, "%s" % url)
            context.finish(True, False, time)
        else:
            context.finish(False, False, time)

    def show_offline_dialog(self):
        return dialogs.report_offline_status(self._widget)

    def get_widget_tree(self):
        return gtk.glade.get_widget_tree(self._widget)

    def present(self):
        self._widget.present()

    def should_present(self):
        if self._widget.window.get_state() is not gtk.gdk.WINDOW_STATE_WITHDRAWN:
            self._widget.hide()
        else:
            self._widget.present()

import os
import gettext
import getopt
import sys
import strawdbus

class Application:
    def __init__(self):
        gnome.program_init(constants.APPNAME.lower(), constants.VERSION)
        self._initialize_config_from_environment()
        gtk.window_set_auto_startup_notification(True)
        self._initialize_gettext()

        config = Config.get_instance()
        config.proxy_config.lookup_host()

        feedlist = FeedList.get_instance()
        feed_categories = FeedCategoryList.get_instance()

        # initialise GUI
        xmlfile = utils.find_glade_file()
        xml = gtk.glade.XML(xmlfile, "straw_main", gettext.textdomain())
        window = xml.get_widget('straw_main')
        self._main_presenter = ApplicationPresenter(view =
                                                    ApplicationView(window))

        if config.first_time:
            libdir = utils.find_data_dir()
            filepath = os.path.join(libdir, "default_subscriptions.opml")
            OPMLImport.import_opml(filepath)
        else:
            feedlist.load_data()
        feed_categories.load_data()

        try:
            itemstore = ItemStore.get_instance(config.straw_dir)
        except ItemStore.ConvertException, ex:
            dialogs.report_error(_("There was a problem while converting the database."),
                                 _("Straw will not behave as expected. You should probably quit now. " +
                                   "The exception has been saved to the file '%s'. Please see the Straw README for further instructions."
                                   ) % ex.reason)
            sys.exit()

        ImageCache.initialize()
        itemstore.start()

        self._main_presenter.view.present()
        PollManager.get_instance().start_polling_loop()

         # set the default icon for the windows
        iconfile = os.path.join(utils.find_image_dir(),"straw.png")
        gtk.window_set_default_icon(gtk.gdk.pixbuf_new_from_file(iconfile))

        strawdbus.start_services()

    def mainloop(self):
        threads = Config.get_instance().use_threads
        if threads: gtk.threads_enter()
        gtk.main()
        if threads: gtk.threads_leave()
        return

    def _initialize_config_from_environment(self):
        threads = os.getenv('STRAW_THREAD_DNS') is not None
        if threads:
            try:
                gtk.threads_init()
            except:
                threads = False
            else:
                threads = True

        config = Config.get_instance()
        config.use_threads = threads
        config.reload_css = os.getenv('STRAW_RELOAD_CSS') is not None
        config.no_etags = os.getenv('STRAW_NO_ETAGS') is not None
        return

    def _initialize_gettext(self):
        import locale
        lname = constants.APPNAME.lower()
        try:
            localedir = utils.find_locale_dir()
            gettext.bindtextdomain(lname, localedir)
            gettext.textdomain(lname)
            gettext.install(lname, localedir, unicode=1)
            gtk.glade.bindtextdomain(lname, localedir)
        except IOError:
            def broken_gettext_workaround(s):
                return s
            __builtins__.__dict__['_'] = broken_gettext_workaround
        locale.setlocale(locale.LC_ALL, '')
        return

    def load_tray(self):
        from Tray import Tray
        tray = Tray()
        tray.connect('button_press_event', self._tray_clicked)

    def _tray_clicked(self, signal, event):
        if event.button == 1:
            self._main_presenter.view.present()
            self._main_presenter.scroll_or_display_next_unread_item()
        else:
            self._main_presenter.view.should_present()
        return



"""  FeedCategoryList.py

Module for feed categories and category subscriptions.
"""
__copyright__ = "Copyright (c) 2002-2005 Free Software Foundation, Inc."
__license__ = """ GNU General Public License

Straw is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

Straw is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA. """

import Feed
from StringIO import StringIO
import Config
import Event
import FeedList
import OPMLImport
import error
import locale
import utils
import types

PSEUDO_ALL_KEY = 'ALL'
PSEUDO_UNCATEGORIZED_KEY = 'UNCATEGORIZED'

class FeedCategoryList(object, Event.SignalEmitter):
    def __init__(self):
        Event.SignalEmitter.__init__(self)
        self.initialize_slots(Event.FeedCategoryChangedSignal,
                              Event.FeedCategorySortedSignal,
                              Event.FeedCategoryAddedSignal,
                              Event.FeedCategoryRemovedSignal,
                              Event.FeedCategoryListLoadedSignal)
        # have to define this here so the titles can be translated
        PSEUDO_TITLES = {PSEUDO_ALL_KEY: _('All'),
                         PSEUDO_UNCATEGORIZED_KEY: _('Uncategorized')}
        self._all_category = PseudoCategory(PSEUDO_TITLES[PSEUDO_ALL_KEY],
                                            PSEUDO_ALL_KEY)
        self._un_category = PseudoCategory(
            PSEUDO_TITLES[PSEUDO_UNCATEGORIZED_KEY], PSEUDO_UNCATEGORIZED_KEY)
        self._user_categories = []
        self._pseudo_categories = (self._all_category, self._un_category)
        self._loading = False
        self._feedlist = FeedList.get_instance()
        self._feedlist.signal_connect(Event.FeedDeletedSignal,
                                self.feed_deleted)
        self._feedlist.signal_connect(Event.FeedCreatedSignal,
                                self.feed_created)
        self._feedlist.signal_connect(Event.FeedsImportedSignal,
                                      self.feeds_imported)

    def load_data(self):
        """Loads categories from config.
        Data format: [[{'title': '', 'subscription': {}, 'pseudo': pseudo key},
        {'id', feed_id1, 'from_subscription': bool} ...], ...]
        """
        cats = Config.get_instance().categories or []
        categorized = {}
        for c in cats:
            head = c[0]
            tail = c[1:]
            pseudo_key = head.get('pseudo', None)
            if pseudo_key == PSEUDO_ALL_KEY:
                fc = self.all_category
            elif pseudo_key == PSEUDO_UNCATEGORIZED_KEY:
                fc = self.un_category
            else:
                fc = FeedCategory(head['title'])
                sub = head.get('subscription', None)
                if sub is not None:
                    fc.subscription = undump_subscription(sub)
            for f in tail:
                # backwards compatibility for stuff saved with versions <= 0.23
                if type(f) is int:
                    fid = f
                    from_sub = False
                else:
                    fid = f['id']
                    from_sub = f['from_subscription']
                feed = self._feedlist.get_feed_with_id(fid)
                if feed is not None:
                    if feed in fc.feeds:
                        error.log("%s (%d) was already in %s, skipping" % (str(feed), fid, str(fc)))
                        continue

                    fc.append_feed(feed, from_sub)
                    categorized[feed] = True
            # User categories: connect pseudos later
            if not pseudo_key:
                fc.signal_connect(Event.FeedCategoryChangedSignal,
                                  self.category_changed)
                self._user_categories.append(fc)
        # just in case we've missed any feeds, go through the list
        # and add to the pseudocategories. cache the feed list of all_category
        # so we don't get a function call (and a list comprehension loop
        # inside it) on each feed. it should be ok here, there are no
        # duplicates in feedlist. right?
        pseudos_changed = False
        all_feeds = self.all_category.feeds
        for f in self._feedlist:
            if f not in all_feeds:
                self.all_category.append_feed(f, False)
                pseudos_changed = True
            uf =  categorized.get(f, None)
            if uf is None:
                self.un_category.append_feed(f, False)
                pseudos_changed = True
        if pseudos_changed:
           self.save_data()
        for cat in self.pseudo_categories:
            cat.signal_connect(
                Event.FeedCategoryChangedSignal, self.pseudo_category_changed)
        self.emit_signal(Event.FeedCategoryListLoadedSignal(self))

    def save_data(self):
        Config.get_instance().categories = [
            cat.dump() for cat in self]

    def pseudo_category_changed(self, signal):
        self.save_data()
        self.emit_signal(signal)

    def category_changed(self, signal):
        if signal.feed is not None:
            uncategorized = True
            for cat in self.user_categories:
                if signal.feed in cat.feeds:
                    uncategorized = False
                    break
            if uncategorized:
                self.un_category.append_feed(signal.feed, False)
            else:
                try:
                    self.un_category.remove_feed(signal.feed)
                except ValueError:
                    pass
        self.save_data()
        self.emit_signal(signal)

    def feed_deleted(self, signal):
        for c in self:
            try:
                c.remove_feed(signal.feed)
            except ValueError:
                pass

    def feed_created(self, signal):
        value = signal.feed
        category = signal.category
        index = signal.index
        if category and category not in self.pseudo_categories:
            if index:
                category.insert_feed(index, value, False)
            else:
                category.append_feed(value, False)
        else:
            self.un_category.append_feed(value, False)
        self.all_category.append_feed(value, False)

    def feeds_imported(self, signal):
        category = signal.category
        feeds = signal.feeds
        from_sub = signal.from_sub

        if category and category not in self.pseudo_categories:
            category.extend_feed(feeds, from_sub)
        else:
            self.un_category.extend_feed(feeds, from_sub)
        self.all_category.extend_feed(feeds, from_sub)
        return

    @property
    def user_categories(self):
        return self._user_categories

    @property
    def pseudo_categories(self):
        return self._pseudo_categories

    @property
    def all_categories(self):
        return self.pseudo_categories + tuple(self.user_categories)

    @property
    def all_category(self):
        return self._all_category

    @property
    def un_category(self):
        return self._un_category

    class CategoryIterator:
        def __init__(self, fclist):
            self._fclist = fclist
            self._index = -1

        def __iter__(self):
            return self

        def _next(self):
            self._index += 1
            i = self._index
            uclen = len(self._fclist.user_categories)
            if i < uclen:
                return self._fclist.user_categories[i]
            elif i < uclen + len(self._fclist.pseudo_categories):
                return self._fclist.pseudo_categories[i - uclen]
            else:
                raise StopIteration

        def next(self):
            v = self._next()
            return v

    def __iter__(self):
        return self.CategoryIterator(self)

    def add_category(self, category):
        category.signal_connect(Event.FeedCategoryChangedSignal,
                                self.category_changed)
        self._user_categories.append(category)
        auxlist = [(x.title.lower(),x) for x in self._user_categories]
        auxlist.sort()
        self._user_categories = [x[1] for x in auxlist]
        self.emit_signal(Event.FeedCategoryAddedSignal(self, category))
        self.save_data()

    def remove_category(self, category):
        for feed in category.feeds:
            category.remove_feed(feed)
        category.signal_disconnect(Event.FeedCategoryChangedSignal,
                                   self.category_changed)
        self._user_categories.remove(category)
        self.emit_signal(Event.FeedCategoryRemovedSignal(self, category))
        self.save_data()

# It might be good to have a superclass FeedCategorySubscription or something
# so we could support different formats. However, I don't know of any other
# relevant format used for this purpose, so that can be done later if needed.
# Of course, they could also just implement the same interface.
class OPMLCategorySubscription(object, Event.SignalEmitter):
    REFRESH_DEFAULT = -1

    def __init__(self, location=None):
        Event.SignalEmitter.__init__(self)
        self.initialize_slots(Event.FeedCategoryChangedSignal,
                              Event.SubscriptionContentsUpdatedSignal)
        self._location = location
        self._username = None
        self._password = None
        self._previous_etag = None
        self._contents = None
        self._frequency = OPMLCategorySubscription.REFRESH_DEFAULT
        self._last_poll = 0
        self._error = None

    @apply
    def location():
        doc = ""
        def fget(self):
            return self._location
        def fset(self, location):
            self._location = location
            self.emit_signal(Event.FeedCategoryChangedSignal(self))
        return property(**locals())

    @apply
    def username():
        doc = ""
        def fget(self):
            return self._username
        def fset(self, username):
            self._username = username
            self.emit_signal(Event.FeedCategoryChangedSignal(self))
        return property(**locals())

    @apply
    def password():
        doc = ""
        def fget(self):
            return self._password
        def fset(self):
            self._password = password
            self.emit_signal(Event.FeedCategoryChangedSignal(self))
        return property(**locals())

    @apply
    def previous_etag():
        doc = ""
        def fget(self):
            return self._previous_etag
        def fset(self, etag):
            self._previous_etag = etag
            self.emit_signal(Event.FeedCategoryChangedSignal(self))
        return property(**locals())

    @apply
    def frequency():
        doc = ""
        def fget(self):
            return self._frequency
        def fset(self, freq):
            self._frequency = freq
            self.emit_signal(Event.FeedCategoryChangedSignal(self))
        return property(**locals())

    @apply
    def last_poll():
        doc = ""
        def fget(self):
            return self._last_poll
        def fset(self, last_poll):
            self._last_poll = last_poll
            self.emit_signal(Event.FeedCategoryChangedSignal(self))
        return property(**locals())

    @apply
    def error():
        doc = ""
        def fget(self):
            return self._error
        def fset(self, error):
            self._error = error
            self.emit_signal(Event.FeedCategoryChangedSignal(self))
        return property(**locals())

    def parse(self, data):
        datastream = StringIO(data)
        entries = OPMLImport.read(datastream)
        contents = [(e.url, e.text) for e in entries]
        updated = contents == self._contents
        self._contents = contents
        if updated:
            self.emit_signal(Event.SubscriptionContentsUpdatedSignal(self))
        return

    @property
    def contents(self):
        return self._contents

    @classmethod
    def undump(klass, dictionary):
        sub = klass()
        sub.location = dictionary.get('location')
        sub.username = dictionary.get('username')
        sub.password = dictionary.get('password')
        sub.frequency = dictionary.get(
            'frequency', OPMLCategorySubscription.REFRESH_DEFAULT)
        sub.last_poll = dictionary.get('last_poll', 0)
        sub.error = dictionary.get('error')
        return sub

    def dump(self):
        return {'type': 'opml',
                'location': self.location,
                'username': self.username,
                'password': self.password,
                'frequency': self.frequency,
                'last_poll': self.last_poll,
                'error': self.error}

def undump_subscription(dictionary):
    try:
        if dictionary.get('type') == 'opml':
            return OPMLCategorySubscription.undump(dictionary)
    except Exception, e:
        error.log("exception while undumping subscription: " + str(e))
        raise

class CategoryMember(object):
    def __init__(self, feed=None, from_sub=False):
        self._feed = feed
        self._from_subscription = from_sub

    @apply
    def feed():
        doc = ""
        def fget(self):
            return self._feed
        def fset(self, feed):
            self._feed = feed
        return property(**locals())

    @apply
    def from_subscription():
        doc = ""
        def fget(self):
            return self._from_subscription
        def fset(self, p):
            self._from_subscription = p
        return property(**locals())

class FeedCategory(list, Event.SignalEmitter):
    def __init__(self, title=""):
        Event.SignalEmitter.__init__(self)
        self.initialize_slots(Event.FeedCategoryChangedSignal,
                              Event.FeedCategorySortedSignal)
        self._title = title
        self._subscription = None

    @apply
    def title():
        doc = ""
        def fget(self):
            return self._title
        def fset(self, title):
            self._title = title
            self.emit_signal(Event.FeedCategoryChangedSignal(self))
        return property(**locals())

    @apply
    def subscription():
        doc = ""
        def fget(self):
            return self._subscription
        def fset(self, sub):
            if self._subscription:
                self._subscription.signal_disconnect(
                    Event.FeedCategoryChangedSignal, self._subscription_changed)
                self._subscription.signal_disconnect(
                    Event.SubscriptionContentsUpdatedSignal,
                    self._subscription_contents_updated)
            if sub:
                sub.signal_connect(Event.FeedCategoryChangedSignal,
                                   self._subscription_changed)
                sub.signal_connect(Event.SubscriptionContentsUpdatedSignal,
                                   self._subscription_contents_updated)
            self._subscription = sub
            self.emit_signal(Event.FeedCategoryChangedSignal(self))
        return property(**locals())

    def read_contents_from_subscription(self):
        if self.subscription is None:
            return
        subfeeds = self.subscription.contents
        sfdict = dict(subfeeds)
        feedlist = FeedList.get_instance()
        current = dict([(feed.location, feed) for feed in self.feeds])
        allfeeds = dict([(feed.location, feed) for feed in feedlist])
        common, toadd, toremove = utils.listdiff(sfdict.keys(), current.keys())
        existing, nonexisting, ignore = utils.listdiff(
            toadd, allfeeds.keys())

        newfeeds = [Feed.Feed.create_new_feed(sfdict[f], f) for f in nonexisting]
        feedlist.extend(self, newfeeds, from_sub=True) # will call extend_feed
        self.extend_feed([allfeeds[f] for f in existing], True)

        for f in toremove:
            index = self.index_feed(allfeeds[f])
            member = self[index]
            if member.from_subscription:
                self.remove_feed(allfeeds[f])
        return

    def _subscription_changed(self, signal):
        self.emit_signal(Event.FeedCategoryChangedSignal(self))

    def _subscription_contents_updated(self, signal):
        self.read_contents_from_subscription()

    def __str__(self):
        return "FeedCategory %s" % self.title

    def __hash__(self):
        return hash(id(self))

    def append(self, value):
        error.log("warning, probably should be using append_feed?")
        list.append(self, value)
        self.emit_signal(Event.FeedCategoryChangedSignal(self, feed=value.feed))

    def append_feed(self, value, from_sub):
        list.append(self, CategoryMember(value, from_sub))
        self.emit_signal(Event.FeedCategoryChangedSignal(self, feed=value))

    def extend_feed(self, values, from_sub):
        list.extend(self, [CategoryMember(v, from_sub) for v in values])
        self.emit_signal(Event.FeedCategoryChangedSignal(self))

    def insert(self, index, value):
        error.log("warning, probably should be using insert_feed?")
        list.insert(self, index, value)
        self.emit_signal(Event.FeedCategoryChangedSignal(self, feed=value.feed))

    def insert_feed(self, index, value, from_sub):
        list.insert(self, index, CategoryMember(value, from_sub))
        self.emit_signal(Event.FeedCategoryChangedSignal(self, feed=value))

    def remove(self, value):
        list.remove(self, value)
        self.emit_signal(Event.FeedCategoryChangedSignal(self, feed=value.feed))

    def remove_feed(self, value):
        for index, member in enumerate(self):
            if member.feed is value:
                del self[index]
                break
        else:
            raise ValueError(value)
        self.emit_signal(Event.FeedCategoryChangedSignal(self, feed=value))

    def reverse(self):
        list.reverse(self)
        self.emit_signal(Event.FeedCategorySortedSignal(self,
                                                        reverse=True))

    def index_feed(self, value):
        for index, f in enumerate(self):
            if self[index].feed is value:
                return index
        raise ValueError(value)

    def _sort_dsu(self, seq):
        aux_list = [(x.feed.title.lower(), x) for x in seq]
        aux_list.sort(lambda a,b:locale.strcoll(a[0],b[0]))
        return [x[1] for x in aux_list]

    def sort(self, indices=None):
        if not indices or len(indices) == 1:
            self[:] = self._sort_dsu(self)
        else:
            items = self._sort_dsu(indices)
            for i,x in enumerate(items):
                list.__setitem__(self, indices[i], items[i])
        self.emit_signal(Event.FeedCategorySortedSignal(self))

    def move_feed(self, source, target):
        if target > source:
            target -= 1
        if target == source:
            return
        t = self[source]
        del self[source]
        list.insert(self, target, t)
        self.emit_signal(Event.FeedCategoryChangedSignal(self))

    def dump(self):
        head = {'title': self.title}
        if self.subscription is not None:
            head['subscription'] = self.subscription.dump()
        return [head] + [
            {'id': f.feed.id, 'from_subscription': f.from_subscription}
            for f in self]

    @property
    def feeds(self):
        return [f.feed for f in self]

    def __eq__(self, ob):
        if isinstance(ob, types.NoneType):
            return 0
        elif isinstance(ob, FeedCategory):
            return self.title == ob.title and list.__eq__(self, ob)
        else:
            raise NotImplementedError

    def __contains__(self, item):
        error.log("warning, should probably be querying the feeds property instead?")
        return list.__contains__(self, item)

class PseudoCategory(FeedCategory):
    def __init__(self, title="", key=None):
        if key not in (PSEUDO_ALL_KEY, PSEUDO_UNCATEGORIZED_KEY):
            raise ValueError, "Invalid key"
        FeedCategory.__init__(self, title)
        self._pseudo_key = key

    def __str__(self):
        return "PseudoCategory %s" % self.title

    def dump(self):
        return [{'pseudo': self._pseudo_key, 'title': ''}] + [
            {'id': f.feed.id, 'from_subscription': False} for f in self]

    def append_feed(self, feed, from_sub):
        assert not from_sub
        FeedCategory.append_feed(self, feed, False)

    def insert_feed(self, index, feed, from_sub):
        assert not from_sub
        FeedCategory.insert_feed(self, index, feed, False)

fclist = None

def get_instance():
    global fclist
    if fclist is None:
        fclist = FeedCategoryList()
    return fclist



# -*- python -*-     -*- encoding: Latin-1 -*-

__copyright__ = "Copyright (c) 2003 Terje Røsten <terjeros@phys.ntnu.no>"

# Extention of distutils for straw.

import sys
import os
import os.path
import re
import glob
import types
import commands

from distutils.core import Command
from distutils.command.build import build
from distutils.command.install import install
from distutils.command.install_lib import install_lib
from distutils.command.install_data import install_data
from distutils.dep_util import newer
from distutils.util import subst_vars
from distutils.dist import Distribution
from distutils.core import setup

class straw_build(build):
    def has_po_files(self):
        return self.distribution.has_po_files()

    def has_desktop_file(self):
        return self.distribution.has_desktop_file()

    sub_commands = []
    sub_commands.extend(build.sub_commands)
    sub_commands.append(('build_mo', has_po_files))
    sub_commands.append(('build_desktop', has_desktop_file))

class build_mo(Command):
    description = 'build binary message catalog'
    user_options = [('build-base=', 'b', 'directory to build to')]
    def initialize_options(self):
        self.build_base = None
        self.force = None

    def finalize_options(self):
        self.set_undefined_options('build',
                                   ('build_base', 'build_base'),
                                   ('force', 'force'))

    def run(self):
        self.announce('Building binary message catalog')
        if self.distribution.has_po_files():
            for mo, po in self.distribution.translations:
                dest = os.path.normpath(self.build_base + '/' + mo)
                self.mkpath(os.path.dirname(dest))
                if not self.force and not newer(po, dest):
                    self.warn("not building %s (up-to-date)" % dest)
                else:
                    import msgfmt
                    msgfmt.make(po, dest)

class build_desktop(Command):
    description = 'builds the straw.desktop file'
    user_options = [('build-base=', 'b', 'directory to build to')]
    def initialize_options(self):
        self.build_base = None
        self.intl_merge = None

    def finalize_options(self):
        if self.intl_merge is None:
            self.intl_merge = commands.getoutput('which intltool-merge')
        if not os.path.exists(self.intl_merge):
            sys.exit('intltool-merge does not exist in your PATH. ' + \
                     'Make sure it exists in your PATH..')
        self.set_undefined_options('build',
                                   ('build_base', 'build_base'))

    def run(self):
        self.announce("Building straw.desktop file....")
        if self.distribution.has_desktop_file():
            for desktop in self.distribution.desktop_file:
                dest = os.path.normpath(os.path.join(self.build_base, 'share/applications'))
                self.mkpath(dest, 1)
                fname = '.'.join(desktop.split('.')[:2]) # straw.desktop.in -> straw.desktop
                cmd = '%s -d -u po %s %s/%s' % (self.intl_merge, desktop,
                                                dest, fname)
                err, val = commands.getstatusoutput(cmd)
                if err:
                    sys.exit('Unable to translate straw.desktop. %s' % err)

class straw_install(install):
    user_options = []
    _user_options = [
        ('sysconfdir=', None, 'specify SYSCONFDIR [default=PREFIX/etc]'),
        ('disable-modules-check', None,
         'do not check that necessary modules is installed'),
        ('enable-modules-check', None,
         'check that necessary modules is installed  [default]'),
        ('disable-schemas-install', None,
         'do not install schema files. Setting the variable '
         'GCONF_DISABLE_MAKEFILE_SCHEMA_INSTALL will prevent this too.'),
        ('enable-schemas-install', None,
         'install schema files. [default]'),
        ('with-gconftool=', None,
         'specify path to the gconftool executable. Can also be set by'
         'the variable GCONFTOOL. [default=gconftool-2]' ),
        ('with-gconf-source=', None,
         'Config database for installing schema files. Can also be set by '
         'the variable GCONF_SCHEMA_CONFIG_SOURCE. Setting to `auto\' is '
         'short for `$(gconftool-2 --get-default-source)\'. '),
        ('with-gconf-schema-file-dir=', None,
         'Directory for installing schema files. Can also be set by the '
         'variable GCONF_SCHEMA_FILE_DIR. [default=SYSCONFDIR/gconf/schemas]'),
        ('with-desktop-file-dir=', None,
         'specify directory for the straw.desktop file [default=PREFIX/share/applications')]
    user_options.extend(install.user_options)
    user_options.extend(_user_options)
    boolean_options = []
    boolean_options.extend(install.boolean_options)
    boolean_options.extend([
        'disable-schemas-install', 'disable-modules-check' ])

    negative_opt = {}
    try:
        negative_opt.update(install.negative_opt)
    except AttributeError:
        pass
    negative_opt.update({'enable-schemas-install' : 'disable-schemas-install',
                         'enable-modules-check' : 'disable-modules-check'})

    def initialize_options(self):
        install.initialize_options(self)
        #self.prefix = '/usr/local'
        # if self.sysconfdir is not a absolute path it will
        # be prefixed by self.prefix
        self.sysconfdir = 'etc'
        self.disable_modules_check = 0
        self.disable_schemas_install = os.environ.get(
            'GCONF_DISABLE_MAKEFILE_SCHEMA_INSTALL', 0)
        self.with_gconftool = os.environ.get(
            'GCONFTOOL', commands.getoutput('which gconftool-2'))
        self.with_gconf_source = os.environ.get(
            'GCONF_SCHEMA_CONFIG_SOURCE', 'auto')
        self.with_gconf_schema_file_dir = os.environ.get(
            'GCONF_SCHEMA_FILE_DIR', None)
        self.with_desktop_file_dir = 'share/applications'

    def finalize_options(self):
        if self.prefix == 'auto':
            cmd = 'pkg-config --variable=prefix libgnome2.0'
            err, val = commands.getstatusoutput(cmd)
            if not err:
                self.prefix = val
            else:
                sys.exit('Cannot find prefix: %s. pkgconfig not installed?'
                         % val)

        self.sysconfdir = os.path.join(self.prefix, self.sysconfdir)
        if self.root:
            self.sysconfdir = os.path.normpath(self.root + '/' + self.sysconfdir)

        self.announce(self.sysconfdir)
        if self.root and self.with_gconf_schema_file_dir:
            self.with_gconf_schema_file_dir = os.path.normpath(
                self.root + '/' + self.with_gconf_schema_file_dir)

        if not self.disable_schemas_install:
            # Sanity check
            if not os.path.exists(self.with_gconftool):
                sys.exit('gconftool-2 executable not found in your path ' + \
                         '- should be installed with GConf')

            if self.with_gconf_source == 'auto':
                cmd = '%s --get-default-source' % self.with_gconftool
                self.with_gconf_source = commands.getoutput(cmd)

            if self.with_gconf_source != 'auto' and self.root:
                self.with_gconf_source = self.with_gconf_source.replace(
                        'xml::','xml::%s' % self.root)
            elif not self.with_gconf_source:
                fmt = 'xml::%s/gconf/gconf.xml.defaults'
                self.with_gconf_source = fmt % self.sysconfdir

        # Run this after we (possibly) have changed prefix.
        install.finalize_options(self)

    def has_modules_check(self):
        return self.distribution.has_modules_check()

    def has_config_files(self):
        return self.distribution.has_config_files()

    def has_po_files(self):
        return self.distribution.has_po_files()

    def has_desktop_file(self):
        return self.distribution.has_desktop_file()

    def has_constants(self):
        return self.distribution.has_constants()

    sub_commands = []
    # Check modules before we start to install files
    sub_commands.append(('install_modules_check', has_modules_check))
    sub_commands.extend(install.sub_commands)
    sub_commands.append(('install_mo', has_po_files))
    sub_commands.append(('install_gconf', has_config_files))
    sub_commands.append(('install_desktop', has_desktop_file))
    sub_commands.append(('install_constants', has_constants))

class install_mo(install_data):
    description = 'install generated binary message catalog'
    def initialize_options(self):
        self.translations = self.distribution.translations
        self.has_po_files = self.distribution.has_po_files
        self.install_dir = None
        self.build_dir = None
        self.skip_build = None
        self.outfiles = []

    def finalize_options(self):
        self.set_undefined_options('build_mo', ('build_base', 'build_dir'))
        self.set_undefined_options('install',
                                   ('install_data', 'install_dir'),
                                   ('skip_build', 'skip_build'))
    def run(self):
        if not self.skip_build:
            self.run_command('build_mo')
        if self.has_po_files():
            for mo, po in self.translations:
                src = os.path.normpath(self.build_dir + '/' + mo)
                if not os.path.isabs(mo):
                    dest =  os.path.normpath(self.install_dir + '/' + mo)
                elif self.root:
                    dest = self.root + mo
                else:
                    dest = mo
                self.mkpath(os.path.dirname(dest))
                (out, _) = self.copy_file(src, dest)
                self.outfiles.append(out)

    def get_outputs (self): return self.outfiles
    def get_inputs (self):  return [ po for mo, po in self.translations ]

class install_gconf(install_data):
    '''
    Same as install_data but using <sysconfdir> instead of
    <prefix>. If <sysconfdir> is not absolute, <prefix> is prefixed
    <sysconfdir>. If the tuple has length 3 and the option in the last
    element is known, the value of the option is used as dest
    directory. On the other hand if option is None, the value of the
    option is set to value of the first element in the tuple.
    '''
    description = 'install generated gconf files'
    user_options = [
        ('sysconfdir', None,
         'specify SYSCONFDIR [default=PREFIX/etc]')]

    def initialize_options(self):
        self.disable_schemas_install = None
        self.sysconfdir = None
        self.with_gconftool = None
        self.with_gconf_source = None
        self.with_gconf_schema_file_dir = None
        self.config_files = self.distribution.config_files
        install_data.initialize_options(self)

    def finalize_options(self):
        install_data.finalize_options(self)
        self.set_undefined_options(
            'install',
            ('disable_schemas_install', 'disable_schemas_install'),
            ('with_gconftool', 'with_gconftool'),
            ('with_gconf_source', 'with_gconf_source'),
            ('sysconfdir','sysconfdir'),
            ('with_gconf_schema_file_dir', 'with_gconf_schema_file_dir'))
        self.option_to_dir()

    def install_schema(self):
        schemas_dir, schemas_file = self.config_files[0]
        schemas_idir = os.path.normpath(self.sysconfdir+'/'+schemas_dir)
        self.mkpath(schemas_idir)
        os.environ['GCONF_CONFIG_SOURCE'] = self.with_gconf_source
        for filepath in schemas_file:
            sfile = filepath.split('/')[-1]
            ofile = os.path.normpath(os.path.join(schemas_idir,sfile))
            self.copy_file(filepath,ofile)
            cmd = '%s --makefile-install-rule %s'\
                  % (self.with_gconftool, ofile)
            err, out = commands.getstatusoutput(cmd)
            if out:
                self.announce(out)
            if err:
                self.warn('Error: installation of gconf schema files failed: %s' % out)

    def run(self):
        if not self.disable_schemas_install:
            self.install_schema()

    def option_to_dir(self):
        data_files = []
        for tup in self.config_files:
            if len(tup) == 2:
                data_files.append(tup)
            elif len(tup) == 3:
                option = tup[2].replace('-','_')
                dest = getattr(self, option, None)
                if dest:
                    data_files.append((dest, tup[1]))
                else:
                    data_files.append((tup[0], tup[1]))
                    dest = os.path.normpath(self.sysconfdir + '/' + tup[0])
                    setattr(self, option, dest)
        self.config_files = data_files

    def get_outputs(self): return []
    def get_inputs(self):  return []

class install_modules_check(Command):
    description = 'check that all necessary Python modules are installed'
    user_options = [
        ('disable-modules-check', None,
         'do not check that necessary modules is installed'),
        ('enable-modules-check', None,
         'check that necessary modules is installed  [default]')]
    boolean_options = []
    boolean_options.append('disable-modules-check')
    negative_opt = {'enable-modules-check' : 'disable-modules-check'}

    def initialize_options(self):
        self.disable_modules_check = None
        self.modules_check = self.distribution.modules_check

    def finalize_options(self):
        self.set_undefined_options(
            'install',
            ('disable_modules_check', 'disable_modules_check'))

    def run(self):
        if self.disable_modules_check:
            self.announce('Modules not checked')
        elif self.modules_check:
            self.modules_check()
            self.announce('All nescessary modules installed')

    def get_outputs(self): return []
    def get_inputs(self):  return []

class install_desktop(install_data):
    description = 'Installs generated desktop file'
    user_options = [
        ('with-desktop-file-dir=', None,
         'specify directory for the straw.desktop file [default=PREFIX/share/applications]')]

    def initialize_options(self):
        self.outfiles = []
        self.build_dir = None
        self.install_dir = None
        self.skip_build = None
        self.with_desktop_file_dir = None

    def finalize_options(self):
        self.set_undefined_options('build_desktop', ('build_base', 'build_dir'))
        self.set_undefined_options('install',('install_data','install_dir'),
                                   ('skip_build', 'skip_build'),
                                   ('with_desktop_file_dir','with_desktop_file_dir'))

    def run(self):
        if not self.skip_build: self.run_command('build_desktop')
        src = os.path.normpath(os.path.join(
            self.build_dir, 'share/applications/straw.desktop'))
        dest = os.path.normpath(self.install_dir + '/' + self.with_desktop_file_dir)
        self.mkpath(dest)
        (out, _) = self.copy_file(src, dest)
        self.outfiles.append(out)

    def get_outputs (self): return self.outfiles
    def get_inputs (self):
        return (os.path.join(self.build_dir, 'share/applications/straw.desktop'))

class translate(Command):
    description = 'update .pot file and merge po files'
    user_options = [('pot', 'p', 'only update the pot file (no merge)')]
    user_options.extend([('dist=', 'd','Merge LANGCODE.po with existing PO template.')])
    boolean_options = ['pot']

    def initialize_options(self):
        self.intl_update = None
        self.pot = 0
        self.dist = None
        self.pot_file = None

    def finalize_options(self):
        if not self.intl_update:
            self.intl_update = commands.getoutput('which intltool-update')
        if not os.path.exists(self.intl_update):
            sys.exit('intltool-update does not exist in your PATH. ' + \
                     'Make sure it exists in your PATH..')

        if not self.pot_file:
            self.distribution.pot_file
        if self.pot and self.dist:
            sys.exit('You can only specify one option at time...\n' + \
                     "'python setup.py translate --help' for more info")

    def run(self):
        # intltool-update needs to be run inside po/
        if os.path.exists('./po/'):
            os.chdir('./po/')
        else:
            sys.exit('po/ directory not found.. not continuing...')

        if self.pot:
            cmd = '%s --pot' % self.intl_update
            err, val = commands.getstatusoutput(cmd)
            if err:
                os.chdir('../')
                sys.exit('Error generating template file %s' % self.pot_file)

        if self.dist:
            cmd = '%s %s' % (self.intl_update, self.dist)
            err, val = commands.getstatusoutput(cmd)
            if err:
                os.chdir('../')
                sys.exit('Error merging %s.po with %s' % (self.dist,self.pot_file))
        os.chdir('../')

class install_constants(install_lib):
    description = "Installs and substitutes file paths and other constants in constants.py.in"
    def initialize_options(self):
        self.build_dir = None
        self.install_lib = None
        self.prefix = None
        self.build_dict = None
        install_lib.initialize_options(self)

    def finalize_options(self):
        install_lib.finalize_options(self)
        self.set_undefined_options('install',
                                   ('install_lib','install_lib'),
                                   ('prefix','prefix'),)
        self.build_dict = {'libdir': os.path.normpath(os.path.join(self.install_lib,'straw')),
                           'datadir': os.path.normpath(os.path.join(self.prefix,'share'))}

    def run(self):
        self.announce("building constants.py...")
        self.fix_constants()
        install_lib.run(self)

    def fix_constants(self):
        if self.distribution.has_constants():
            for fname,vars_dict in self.distribution.constants:
                vars_dict.update(self.build_dict)
                f = open(fname,'r')
                cdata = f.read()
                f.close()
                sfname = '.'.join(fname.split('.')[:2]) # foo.py.in -> foo.py
                f = open(os.path.normpath(os.path.join(self.build_dir, 'straw', sfname)),'w')
                f.write(subst_vars(cdata, vars_dict))
                f.close()
            self.announce("... done.")

class StrawDistribution(Distribution):
    def __init__(self, attrs = None):
        self.modules_check = 0
        self.msg_sources = None
        self.pot_file = None
        self.translations = []
        self.config_files = []
        self.desktop_file = []
        self.constants = []
        self.ext_modules = []
        Distribution.__init__(self, attrs)
        self.cmdclass = {
            'install_modules_check' : install_modules_check,
            'build' : straw_build,
            'build_mo' : build_mo,
            'translate' : translate,
            'build_desktop': build_desktop,
            'install' : straw_install,
            'install_mo' : install_mo,
            'install_desktop' : install_desktop,
            'install_constants': install_constants,
            'install_gconf' : install_gconf}

    def has_po_files(self): return len(self.translations) > 0
    def has_modules_check(self): return isinstance(self.modules_check, types.FunctionType)
    def has_config_files(self): return len(self.config_files) > 0
    def has_desktop_file(self): return len(self.desktop_file) > 0
    def has_constants(self): return len(self.constants) > 0

def setup(**kwds):
    from distutils.core import setup
    kwds['distclass'] = StrawDistribution
    setup(**kwds)



""" PollManager.py

Module for polling the feeds.
"""
__copyright__ = "Copyright (c) 2002-4 Juri Pakaste <juri@iki.fi>"
__license__ = """ GNU General Public License

Straw is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

Straw is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA. """

import Config
import FeedList
import Event
from MainloopManager import MainloopManager
import MessageManager
import NetworkConstants
import URLFetch
import Feed
import FeedCategoryList
import SummaryParser
import utils
import time
import socket
import error

class PollStopper(object, Event.SignalEmitter):
    def __init__(self, stopper, object):
        Event.SignalEmitter.__init__(self)
        self.initialize_slots(Event.PollingStoppedSignal)
        self._stopper = stopper
        self._object = object

    stopper = property(lambda self: self._stopper)
    object = property(lambda self: self._object)

    def stop(self):
        if self._stopper is not None:
            self._stopper()
            self._stopper = None
            self.emit_signal(Event.PollingStoppedSignal(self))

class PollManager:
    def __init__(self):
        self._config = Config.get_instance()
        self._mmgr = MessageManager.get_instance()
        self._feedlist = FeedList.get_instance()

        self._config.signal_connect(Event.PollFrequencyChangedSignal, self.poll_changed)
        self._config.signal_connect(Event.OfflineModeChangedSignal, self.offline_changed)
        self._feedlist.signal_connect(Event.FeedsChangedSignal, self.feeds_changed)

        self._is_offline = self._config.offline
        self._poll_frequency = self._config.poll_frequency
        self._last_poll = self._config.last_poll
        self._feeds = self._feedlist.flatten_list()
        self._pollers = {}

    def poll_changed(self, signal):
        """ called when global poll frequency has changed"""
        self._poll_frequency = signal.value

    def offline_changed(self, signal):
        self._is_offline = self._config.offline

    def feeds_changed(self, *args):
        self._feedlist = FeedList.get_instance()
        self._feeds = self._feedlist.flatten_list()

    def start_polling_loop(self):
        mlmgr = MainloopManager.get_instance()
        mlmgr.set_repeating_timer(NetworkConstants.POLL_INTERVAL, self.maybe_poll)

    ##########################
    # Why both pollcontext and self._pollers?
    # self._pollers is there to make sure that there are no simultaneous
    # pollers on the same object. So the user doesn't accidentally start
    # a poller for an object being polled.
    # pollcontext is there so that we can avoid starting pollers many
    # times - even though they wouldn't overlap - on the same object within
    # one polling run.

    # because maybe_poll is used as an idle handler, it must always return
    # True or else it's removed from the handler list
    def maybe_poll(self):
        self._maybe_poll()

    def _maybe_poll(self):
        now = int(time.time())
        if self._is_offline:
            return True
        try:
            context = {}
            config_pf = self._config.poll_frequency
            feed_use_default = Feed.Feed.DEFAULT
            self.poll([feed for feed, timediff, fpf in
                       [(feed, now - feed.last_poll, feed.poll_frequency)
                        for feed in self._feeds]
                       if ((timediff > fpf > 0) or
                           (fpf == feed_use_default and
                            timediff > config_pf > 0))], context)


            self.poll_categories([cat for cat, sub, timediff in
            [(cat, cat.subscription, now - cat.subscription.last_poll)
             for cat in FeedCategoryList.get_instance().all_categories
             if cat.subscription is not None]
            if ((timediff > sub.frequency > 0) or
                (sub.frequency == sub.REFRESH_DEFAULT and
                 timediff > config_pf > 0))], context, False)
        except:
            error.log_exc("Caught an exception while polling")
        URLFetch.connection_manager.poll(NetworkConstants.POLL_TIMEOUT)
        return True

    def poll(self, obj, pollcontext = None):
        """ obj must be a list of Feed objects"""
        if pollcontext is None:
            pollcontext = {}
        stoppers = []
        for f in obj:
            sf = None
            if not self._pollers.has_key(f) and not pollcontext.has_key(f):
                self._pollers[f] = True
                pollcontext[f] = True
                sf = FeedPoller(f).poll()
            stoppers.append(sf)
        return stoppers

    def poll_categories(self, cats, pollcontext = None, feeds = True):
        """Polls categories and, if feeds is True, the feeds
        associated with them.  Returns a list of list containing
        PollStopper objects.  poll_categories([c1, c2, c3]) -> [[sc1,
        sc1f1, sc1f2 ...], [sc2, sc2f1, sc2f2 ...], ...] where cN is a
        category, scN is a PollStopper for cN, scNfM is a stopper for
        a feed in cN.  If no polling was started for that cN or cNfM,
        in its place is None.  """
        stoppers = []
        if pollcontext is None:
            pollcontext = {}
        for c in cats:
            category_stoppers = []
            sc = None
            if c.subscription is not None and not self._pollers.has_key(c) and not pollcontext.has_key(c):
                self._pollers[c] = True
                pollcontext[c] = True
                sc = CategoryPoller(c, pollcontext).poll()
            category_stoppers.append(sc)
            if feeds:
                for f in c.feeds:
                    sf = None
                    if not self._pollers.has_key(f) and not pollcontext.has_key(f):
                        self._pollers[f] = True
                        pollcontext[f] = True
                        sf = FeedPoller(f).poll(), f
                    category_stoppers.append(sf)
            stoppers.append(category_stoppers)
        return stoppers

    def poll_done(self, obj):
        if self._pollers.has_key(obj):
            del self._pollers[obj]

class FeedPoller:
    def __init__(self, feed):
        self._feed = feed
        self._mmgr = MessageManager.get_instance()

    def poll(self):
        MainloopManager.get_instance().call_pending()

        url, user, pw = self._feed.access_info
        parsed = None
        headers = {}
        config = Config.get_instance()
        if not config.no_etags and self._feed.previous_etag is not None:
            headers['If-None-Match'] = self._feed.previous_etag

        self._feed.last_poll = int (time.time())
        ps = None
        try:
            try:
                stopper = URLFetch.connection_manager.request(
                    url, self, headers, user, pw,
                    priority=NetworkConstants.PRIORITY_RSS)
                ps = PollStopper(stopper, self._feed)
            except Exception, e:
                error.log_exc("Caught an exception while polling")
                self.http_failed(e)
            else:
                self._feed.router.start_polling()
                self._feed.router.stopper = ps
        finally:
            return ps

    def http_results(self, status, header, data):
        MainloopManager.get_instance().call_pending()
        try:
            try:
                import pdb
                #pdb.set_trace()
                self._feed.poll_done()
                err = ""
                read_data = True
                if status is None:
                    err = _("No data")
                elif status[1] == 304:
                    self._feed.router.route_no_data()
                    read_data = False
                elif status[1] == 410 or status[1] == 404:
                    err = _("Unable to find the feed (%s: %s)") % (
                        status[1], status[2].strip())
                elif status[1] == 401:
                    err = _("Invalid username and password.")
                elif status[1] > 299:
                    err = _("Updating feed resulted in abnormal status '%s' (code %d)") % (
                        status[2].strip(), status[1])

                if err:
                    self._feed.router.set_error(err)
                elif read_data:
                    try:
                        parsed = SummaryParser.parse(data, self._feed)
                    except Exception, e:
                        error.log_exc("exception in summaryparser")
                        self._feed.router.set_error(
                            _("An error occurred while processing feed: %s") %
                            str(e))
                    else:
                        self._feed.router.route_all(header, parsed)
                self._mmgr.post_message(_("Updating %s done.") % self._feed.title)
            except Exception, ex:
                error.log_exc("error while parsing results")
                self._feed.router.set_error(str(ex))
        finally:
            get_instance().poll_done(self._feed)

    def http_failed(self, exception):
        try:
            if isinstance(exception, socket.error):
                self._feed.router.route_no_data()
            else:
                self._feed.router.set_error(str(exception))
            self._mmgr.post_message(_("Updating %s failed") % self._feed.title)
            self._feed.poll_done()
        finally:
            get_instance().poll_done(self._feed)

    def http_permanent_redirect(self, location):
        (oldloc, u, p) = self._feed.access_info
        self._feed.access_info = (location, u, p)

    def operation_stopped(self):
        self._feed.router.route_no_data()
        self._feed.poll_done()
        get_instance().poll_done(self._feed)

class CategoryPoller:
    def __init__(self, category, pollcontext):
        self._category = category
        self._mmgr = MessageManager.get_instance()
        self._pollcontext = pollcontext

    def poll(self):
        headers = {}
        sub = self._category.subscription
        if sub is None or sub.location is None or len(sub.location) == 0:
            return None
        sub.last_poll = int(time.time())
        if sub.previous_etag is not None:
            headers['If-None-Match'] = sub.previous_etag
        ps = None
        try:
            try:
                stopper = URLFetch.connection_manager.request(
                    sub.location, self, headers, sub.username, sub.password,
                    priority=NetworkConstants.PRIORITY_RSS)
                ps = PollStopper(stopper, self._category)
            except Exception, e:
                error.log_exc("Caught an exception while polling category")
        finally:
            return ps

    def http_results(self, status, header, data):
        sub = self._category.subscription
        try:
            try:
                err = None
                if status is None:
                    err = _("No data (%s)" % status[1])
                elif status[1] == 410 or status[1] == 404:
                    err = _("Unable to find the category (%s: %s)") % (
                        status[1], status[2].strip())
                elif status[1] == 401:
                    err = _("Invalid username and password.")
                elif status[1] > 299:
                    err = _("Updating category resulted in abnormal status '%s' (code %d)") % (status[2].strip(), status[1])

                if err is not None:
                    self._category.subscription.error = err
                else:
                    sub.parse(data)
                    old_feeds = self._category.feeds
                    self._category.read_contents_from_subscription()
                    common, inold, innew = utils.listdiff(
                        old_feeds, self._category.feeds)
                    if len(innew) > 0:
                        get_instance().poll(innew, self._pollcontext)
            except Exception, e:
                self._category.subscription.error = str(e)
        finally:
            get_instance().poll_done(self._category)

    def http_failed(self, exception):
        self._category.subscription.error = str(exception)
        get_instance().poll_done(self._category)

    def http_permanent_redirect(self, location):
        error.logparam(locals(), "location")

    def operation_stopped(self):
        get_instance().poll_done(self._category)

pollmanager_instance = None

def get_instance():
    global pollmanager_instance
    if pollmanager_instance is None:
        pollmanager_instance = PollManager()
    return pollmanager_instance



""" FeedDataRouter.py

Module for routing feed data from the poller
"""
__copyright__ = "Copyright (c) 2002-2005 Free Software Foundation, Inc."
__license__ = """
Straw is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

Straw is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA. """


import Event
import ImageCache

class FeedDataRouter(object):
    def __init__(self, feed):
        ImageCache.cache.signal_connect(Event.ImageUpdatedSignal,
                                              self.image_updated)
        self._feed = feed
        self._stopper = None
        # stuff we are potentially waiting for
        self._wait_data = 0
        self._wait_images = {}

    # public parts
    def start_polling(self):
        self._wait_data = 1
        self.update_process_status()

    def route_no_data(self):
        self._wait_data = 0
        self._feed.error = None
        self._stopper = None
        self.update_process_status()

    def route_all(self, header, parsed):
        needed_loading = self._feed.load_contents()
        self.route_data(header, parsed)
        self.route_extras()
        if needed_loading:
            self._feed.unload_contents()

    def route_data(self, header, parsed):
        if header:
            etag = header.getheader('etag', None)
            if etag:
                etag = etag.strip()
                self._feed.previous_etag = etag
        self._feed.channel_title = parsed.title
        self._feed.channel_description = parsed.description
        self._feed.channel_copyright = parsed.copyright
        self._feed.channel_link = parsed.link
        self._feed.channel_creator = parsed.creator
        if len(parsed.items):
            parsed.items.reverse()
            self._feed.add_items(parsed.items)
        self._feed.error = None
        self._wait_data = 0
        self._feed.poll_done()

    def route_extras(self):
        try:
            for i in self._feed.items:
                for name in i.image_keys():
                    image = i.get_image(name)
                    if image.status == image.WAITING:
                        self._wait_images[image.url] = 1
        except Exception, ex:
            # if we don't get the images, we can probably still show
            # the text
            pass
        self.update_process_status()

    def set_error(self, err):
        self._wait_data = 0
        self._feed.error = err
        self.update_process_status()
        return

    def set_stopper(self, stopper):
        self._stopper = stopper
        stopper.signal_connect(Event.PollingStoppedSignal,
                               self._polling_stopped)
    def get_stopper(self): return self._stopper
    stopper = property(get_stopper, set_stopper)

    def poll_done(self):
        self._feed.poll_done()
        self._stopper = None

    def stop_polling(self):
        if self._stopper:
            self._stopper.stop()
            self._stopper = None
        for image in self._wait_images.keys():
            ImageCache.cache.stop_transfer(image)
            del self._wait_images[image]
        self.update_process_status()
        self._feed.poll_done()

    def _polling_stopped(self, signal):
        self.poll_done()

    # private parts

    def image_updated(self, signal):
        if signal.url in self._wait_images:
            del self._wait_images[signal.url]
        self.update_process_status()

    def update_process_status(self):
        length_images = len(self._wait_images)
        if self._wait_data or length_images:
            self._feed.process_status = self._feed.STATUS_POLLING
        else:
            self._feed.process_status = self._feed.STATUS_IDLE



#! /usr/bin/env python
# -*- coding: iso-8859-1 -*-
#
# Copyright Terje Røsten <terjeros@phys.ntnu.no> Nov. 2003.
# 
'''Merge two Uniforum style .po files together.

This is a implementation (not complete) in Python of the GNU
msgmerge(1) program. It can be used on the command line (or as a Python
module).

Usage: msgmerge.py [OPTIONS] def.po ref.pot

The def.po file is an existing PO file with translations. The ref.pot
file is the last created PO file with up-to-date source references but
old translations, or a PO Template file.

Options:
  -U, --update           update def.po,
                         do nothing if def.po is already up to date.
  -o, --output-file=FILE write output to file FILE. Output is written to
                         stdout if set to - or if the option is not present.
  -D, --docstrings       don\'t remove docstring flag.
  -h, --help             display help text and exit.
  -V, --version          display version and exit.
  -q, --quiet, --silent  suppress progress indicators.
'''
from __future__ import generators

if not __name__ == '__main__':
    __doc__ += '''\

When used as module the interesting functions are merge() and
merge_dir().

The merge() function does the same as the command line version, and
the arguments are as follows. The first argument is the def.po file,
then the ref.pot file. The third argument controls whether do work in
update mode or not, then the next argument sets the output file. Set
the next argument to False to remove docstring flags. The last
argument can be used to suppress progress indicators. The default is
to work in update mode with progress indicators.

Example:
 merge("def.po", "ref.pot")
  merge the files def.po and ref.pot and write output to def.po if
  there are any changes.
 merge("def.po", "red.pot", docstrings = False, verbose = False,
       update = False, outfile = "-")
  merge the files def.po and ref.pot and write output to stdout,
  remove docstring flag and be quiet.

The merge_dir() function is useful when merging a directory of po
files. The only required argument is the name of the directory with po
files and the pot file. It will use simple glob to find the files. The
second argument can be used to specify the pot file (in the
directory). Third argument is a list of po files (then globbing will
not be used) and the next argument is list of filename to exclude. The
last argument can be used to suppress progress indicators. Docstring
flag will not be removed.

Example:
 merge_dir("po")
  merge (and update) all po files in directory po with the single pot
  file in the same directory.

The module raises the MsgmergeError exception in case of error.
'''
__version__ = '0.1'
name = 'msgmerge.py'

__all__ = [ 'merge', 'merge_dir', 'MsgmergeError' ]

import sys
import re
import string
import getopt
import difflib
import glob
import os.path

try:
    True, False
except NameError:
    True, False = 1, 0

class Msgs:
    '''Class to hold information about messages.'''
    width = 80
    file = ''
    def __init__(self, msgid, msgstr, flag, lno, entry, **kwds):
        self.id = msgid
        self.str = msgstr
        self.cmt = kwds.get('cmt', '')
        self.ref = kwds.get('ref', '')
        self.autocmt = kwds.get('autocmt', '')
        self.flag = flag
        self.entry = entry
        self.lno = lno
        self.count = 0
    def wash(self):
        self.id = wash(self.id, width = self.width,
                       filename = self.file, lno = self.lno)
        self.str = wash(self.str, 'msgstr', width = self.width,
                        filename = self.file, lno = self.lno)
    def used(self):
        self.count += 1
    def get_clean_id(self):
        return self.id.replace('msgid "','', 1)
    def obsolete(self):
        self.width -= len('#~ ')
        self.wash()
        t = [ '#~ %s\n' % s for s in self.id.splitlines() ]
        self.id = ''.join(t)
        t = [ '#~ %s\n' % s for s in self.str.splitlines() ]
        self.str = ''.join(t)

class Options:
    '''Class to hold options'''
    def __init__(self, cmdline = False, **kwds):
        if not cmdline:
            self.update = kwds.get('update', True)
            self.outfile = kwds.get('outfile', '-')
            self.docstrings = kwds.get('docstrings', True)
            self.verbose = kwds.get('verbose', False)
        else:
            self.update = False
            self.outfile = False
            self.docstrings = False
            self.verbose = True

class MsgmergeError(Exception):
    '''Exception class for msgmerge'''
    def __init__(self, args):
        self.args = args

def gen(lines):
    '''
    Generator which returns a line (with the obsolete prefix removed)
    from the list of lines in <lines>, the line number is also
    returned.
    '''
    lno = 0
    for l in lines:
        lno += 1
        yield l.replace('#~ ', '', 1), lno
    yield l, lno

def slurp(s, g, sign):
    '''
    The string returned from iterator <g>\'s next() method is added to
    the string <s> if string returned is beginning with the string
    <sign>. The return value is the first returned string which do not
    start with <sign>, the line number, the iterator <g> and the
    (possibly) updated string <s>.
    '''
    l, lno = g.next()
    while l.startswith(sign) or (sign == '# ' and l.strip() == '#'):
        s += l
        l, lno = g.next()
    return l, lno, g, s

def splitted_fit(chunk, line, width, break_always, break_after_space):
    '''
    Check if string <chunk> can be splitted by newline to fit into
    string <line> with width smaller than <width>. The return value is
    a tuple where the first element is the part of chunk which fits
    and the second element is the rest of chunk.
    '''
    ret = '', chunk
    l = len(chunk)
    for i in range(l - 1, -1, -1):
        if chunk[i] in break_always and len(chunk[0:i] + line) <= width:
            ret = chunk[0:i], chunk[i:]
            break
        elif chunk[i] in break_after_space and i and chunk[i-1].strip() == '':
            ret = chunk[0:i], chunk[i:]
            break
        elif chunk[i] == '\\' and len(chunk[i:]) > 1 and chunk[i+1] == '"' \
             and len(chunk[0:i] + line) <= width:
            ret = chunk[0:i], chunk[i:]
            break
    return ret

def wrap(msg, width):
    '''
    Accept a list <msg> of strings to wrap, each string is wrapped to
    width <width> and surrounded with a pair of ". The return value is
    a string with these wrapped strings joined together with newlines.
    '''
    if msg.isspace() or not msg:
        return '"%s"' % msg

    # \ and " is here, but " is special in po files.
    break_always = '$%+({['
    # XXX what about: « © » ¦ § etc?
    break_after_space = '_-=^`~\'<|>&*#@'
    enders = '.:,;!?/])}|%-'
    extra = string.punctuation
    for c in enders:
        extra = extra.replace(c, '')
    escaped = { 'enders' : re.escape(enders),
                'extra'  : re.escape(extra) }
    regex = r'([\w%(extra)s]*[\s%(enders)s)]+[\s%(enders)s]*)' % escaped
    r = re.compile(regex, re.UNICODE)
    msg = [ m for m in r.split(msg) if not m == '']

    lines = []
    line = msg.pop(0)
    
    # Handle \n on end of line
    if len(msg) > 1 and msg[-1] == 'n' and len(msg[-2]) > 0 \
           and msg[-2][-1] == '\\':
        msg[-2] += msg[-1]
        msg.pop()
    # Do not allow a single \n on a line
    if len(msg) > 2 and msg[-1] == '\\n':
        msg[-2] += msg[-1]
        msg.pop()

    for m in msg:
        if len(line) > width or len(m) > width or len(line + m) > width:
            fit, rest = splitted_fit(m, line, width, break_always,
                                     break_after_space)
            line += fit
            lines.append(line)
            line = rest
        else:
            line += m
    lines.append(line)
    lines = [ '"%s"' % l for l in lines ]
    return '\n'.join(lines)

def normalize(lines):
    '''
    Normalize <lines>: e.g "\n\nText\n\n" becomes:
    "\n"
    "\n"
    "Text\n"
    "\n"
    '''
    if  0 < lines.find('\\n') < len(lines) - 3:
        if lines[-3:] == '\\n"':    
            lines = lines[:-3].replace('\\n','\\n"\n"').replace('""\n','') \
                    + '\\n"'
        else:
            lines = lines.replace('\\n','\\n"\n"').replace('""\n','')
    return lines

def wash(msg, idx = 'msgid', width = 80, **kwds):
    '''
    Do washing on the msgstr or msgid fields. Wrap the text to fit in
    width <width>. <msg> is a list of lines that makes up the field.
    <idx> indicate msgid or msgstr, <width> holds the width. <filename>
    and <lno> (line number) is picked up from <kwds>.
    Returns the washed field as a string.
    '''
    msg = normalize(msg)
    lines = msg.splitlines()
    size = len(lines)
    if size > 1 or len(msg) > width:
        washed = []
        # The first line is special
        m = re.match('^%s "(.*)"$' % (idx, ), lines[0])
        if not m:
            print lines[0]
            kwds['lno'] -= size + 1            
            raise MsgmergeError('parse error: %(filename)s:%(lno)s.'
                                % kwds)
        washed.append(m.group(1))
        if m.group(1).endswith(r'\n'):
            washed.append('')
        i = 0
        for line in lines[1:]:
            m = re.match('^"(\s*.*)"$', line)
            i += 1
            if not m:
                print line
                kwds['lno'] -= size - i + 1
                raise MsgmergeError('parse error: %(filename)s:%(lno)s.'
                                    % kwds)
            washed[-1] += m.group(1)
            if m.group(1).endswith(r'\n'):
                washed.append('')
        if washed[0] == '':
            washed.pop(0)
        if washed[-1] == '':
            washed.pop()
    
        washed = [ wrap(w, width - 3) for w in washed ] # " and \n removed.

        # One line or multiline
        if len(washed) == 1 and len('%s %s\n' % (idx, washed[0])) < width:
            washed = '%s %s\n' % (idx, washed[0])
        else:
            washed = '%s ""\n%s\n' % (idx, '\n'.join(washed))
    else:
        washed = msg

    return washed

def parse(filename, entry):
    '''
    Parse po or pot file with name <filename>. Set the variable
    <entry> to msgid/msgstr to indicate pot/po file.  The return value
    is a dict with msgid (washed) as key and Msgs instances as
    values.
    '''
    lines = io(filename).readlines()
    Msgs.file = filename
    messages = {}
    last = len(lines)
    g = gen(lines)            
    cmt = autocmt = ref = flag = ''
    msgid = False
    lno = 0
    while not lno == last:    
        l, lno = g.next()
        if l.startswith('# '):
            l, lno, g, cmt  = slurp(l, g, '# ')
        if l.startswith('#.'):
            l, lno, g, autocmt = slurp(l, g, '#.')
        if l.startswith('#:'):
            l, lno, g, ref = slurp(l, g, '#:')
        if l.startswith('#,'):
            l, lno, g, flag = slurp(l, g, '#,')
        if l.startswith('msgid'):
            l, lno, g, msgid = slurp(l, g, '"')
        if l.startswith('msgstr'):
            l, lno, g, msgstr = slurp(l, g, '"')

        if not lno == last and not l.strip() == '':
            raise MsgmergeError('parse error: %s:%s.' % (filename, lno))

        if msgid and entry == 'msgstr':
            idx = wash(msgid, filename = filename, lno = lno)
            messages[idx] = Msgs(msgid, msgstr, flag, lno, entry, cmt = cmt)
            msgid = False; msgstr = cmt = autocmt = ref = flag = ''
        elif msgid and entry == 'msgid':
            idx = wash(msgid, filename = filename, lno = lno)
            messages[idx] = Msgs(msgid, msgstr, flag, lno, entry,
                                 autocmt = autocmt, ref = ref)
            msgid = False; msgstr = cmt = autocmt = ref = flag = ''

    for m in messages.values():
        m.wash()
    return messages

def fuzzy_match(pot, defs):
    '''
    Try to find the best difflib match (with ratio > 0.6) between
    id of Msgs object <pot> and Msgs in the dict <defs>.
    Return value is the Msgs object in <defs> with highest ratio,
    False is returned if no suitable Msgs is found.
    '''
    limit = 0.6
    l, po = limit - 0.01, False
    s = difflib.SequenceMatcher(lambda x: x == ' "', '', pot.get_clean_id())
    len2 = len(pot.get_clean_id())   
    for candidate in defs.values():
        if candidate.str == 'msgstr ""\n':       # Empty translation
            continue
        if candidate.id == 'msgid ""\n':         # Empty msgid (header)
            continue
        len1 = len(candidate.get_clean_id())
        if len2 > 2 * len1 or len1 > 1.5 * len2: # Simple and fast tests first
            continue
        s.set_seq1(candidate.get_clean_id())
        if s.quick_ratio() < l:
            continue
        r = s.ratio()                            # This is expensive
        if r > l:
            l, po = r, candidate
    return po

def flags(po, pot, fuzzy = False, obs = False):
    '''
    Create flag field from flag field in Msgs objects <po> and
    <pot>. When <fuzzy> is true <po>\'s flags are ignored and the
    fuzzy flag is added. If <obs> is set then most flags but fuzzy are
    removed. If the global variable option.docstrings is set then
    docstring flags will not be removed. The return value is a string
    which holds the combined flag.
    '''
    global option
    flag = ''
    if po.flag or pot.flag or fuzzy:
        if not fuzzy:
            flag = '%s, %s' % (po.flag.strip(), pot.flag.strip())
        else:
            flag = '%s, %s' % ('#, fuzzy', pot.flag.strip())
        flag = flag.split(', ')
        fl = {}
        flag = [fl.setdefault(f, f) for f in flag if f not in fl and f]
        if not option.docstrings:
            try:
                flag.remove('docstring')
            except ValueError:
                pass
        if obs:
            removes = ['c-format', 'python-format', 'docstring']
            for remove in removes:
                try:
                    flag.remove(remove)
                except ValueError:
                    pass
        # Put fuzzy first
        if 'fuzzy' in flag and not flag.index('fuzzy') == 1:
            i = flag.index('fuzzy')
            flag[1], flag[i] = flag[i], flag[1]

        if len(flag) == 1:
            flag = ''
        else:
            flag = ', '.join(flag) + '\n'
    return flag

def add(pot, po, fuzzy = False):
    '''
    Build a new entry from the Msgs objects <pot> and <pot>. If
    <fuzzy> is true, <po>\'s flag field is ignored (in
    flags()). Returns a multiline string with a up to date entry.
    '''
    msg = []
    msg.append(po.cmt)
    msg.append(pot.autocmt)
    msg.append(pot.ref)
    msg.append(flags(po, pot, fuzzy = fuzzy))
    msg.append(pot.id)
    msg.append(po.str)
    return ''.join(msg)

def header(pot, defs):
    '''
    Update date in header entry. Returns the updated header entry.
    '''
    try:
        [po] = [ d for d in defs.values() if d.id == 'msgid ""\n' ]
    except ValueError:
        raise MsgmergeError('Error: did not find header in po file.')

    r = re.compile(r'(.*^"POT-Creation-Date:\s+)(.*?)(\\n"$.*)',
                   re.MULTILINE | re.DOTALL)
    m = r.match(pot.str)
    if not m:
        raise MsgmergeError(
            'Error: did not find POT-Creation-Date field in pot file.')

    subs = '\\1%s\\3' % m.group(2)
    po.str, count = r.subn(subs, po.str)
    if not count == 1:
        raise MsgmergeError(
            'Error: did not find POT-Creation-Date field in po file.')
    return po

def match(defs, refs):
    '''
    Try to match Msgs objects in <refs> with Msgs objects in
    <defs>. The return value is a list with po entries.
    '''
    global option
    matches = []
    empty = Msgs('msgid ""\n', 'msgstr ""\n', '', -1, 'str')
    deco = [(r.lno, r) for r in refs.values() ]
    deco.sort()
    po = header(deco.pop(0)[1], defs)       # Header entry
    matches.append(add(empty, po))
    po.used()
    sorted = [ a[1] for a in deco ]
    for pot in sorted:
        if option.verbose:
            sys.stderr.write('.')
        po = defs.get(pot.id, False)        # Perfect match
        if po:
            matches.append(add(pot, po))
            po.used(); pot.used()
            continue
        po = fuzzy_match(pot, defs)         # Fuzzy match
        if po:
            matches.append(add(pot, po, fuzzy = True))
            po.used(); pot.used()
            continue
        matches.append(add(pot, empty))     # No match

    obsolete(defs, matches)
    return matches

def obsolete(defs, matches):
    '''Handle obsolete translations.'''
    deco = [ (d.lno, d) for d in defs.values() if
             d.count == 0 and not d.str == 'msgstr ""\n' ]
    deco.sort()
    empty = Msgs('msgid ""\n', 'msgstr ""\n', '', -1, 'str')
    obs = [ o[1] for o in deco ]
    for o in obs:
        o.flag = flags(o, empty, obs = True) 
        o.obsolete()
        matches.append('%s%s%s' % (o.flag, o.id, o.str))

def help():
    '''Print help text and exit.'''
    print __doc__
    sys.exit(0)

def cmdline():
    '''Parse options and arguments from command line.'''
    advice = 'Try `%(name)s --help\' for more information.'
    try:
        long_opt = ['help', 'version', 'update', 'output-file=',
                    'quiet', 'silent', 'docstrings']
        opts, args = getopt.getopt(sys.argv[1:], 'hVUo:qD', long_opt)
    except getopt.error, msg:
        print '%s: %s\n%s' % ('%(name)s', msg, advice) % globals()
        sys.exit(1)
        
    option = Options(cmdline = True)
    for opt, arg in opts:
        if opt in ['-h', '--help']:
            help()
        elif opt in ['-V', '--version']:
            print '%(name)s %(__version__)s' % globals()
            sys.exit(0)
        elif opt in ['-o', '--output-file']:
            option.outfile = arg
        elif opt in ['-U', '--update']:
            option.update = True
        elif opt in ['-q', '--silent', '--quiet']:
            option.verbose = False
        elif opt in ['-D', '--docstrings']:
            option.docstrings = True
            
    # Sanity checks
    warn = False
    if option.update and option.outfile:
        warn = '--update and --output-file are mutually exclusive.'
    if len(args) == 0:
        warn = 'no input files given.'
    elif len(args) == 1 or len(args) > 2:
        warn = 'exactly 2 input files required.'
    if warn:
        print '%s: %s\n%s' % ('%(name)s', warn, advice) % globals()
        sys.exit(1)

    if option.update:
        option.outfile = args[0]
    elif not option.outfile:
        option.outfile = '-'

    defs, refs = args

    try:
        merge(defs, refs, option = option)
    except MsgmergeError, err:
        print '%(name)s: ' % globals() + '%s' % err
        sys.exit(1)

def io(iofile, mode = 'r'):
    '''Wrapper around open().'''
    try:
        fd = open(iofile, mode)
    except IOError, msg:
        raise MsgmergeError('error while opening file: %s: %s.' %
                            (msg[1], iofile))
    return fd

def changes(new, old):
    return cmp(''.join(old), '\n'.join(new))

def write(matches, outfile):
    '''Write the list <matches> to file <outfile>'''
    if not outfile == '-':
        fd = io(outfile, 'w')
    else:
        fd = sys.stdout
    fd.write('\n'.join(matches))
    
def merge(def_file, ref_file, update = True, outfile = '-',
          docstrings = True, verbose = True, **kwds):
    '''
    Merge po file <def_file> with pot file <ref_file> . If <update> is
    set to True then only update if there are changes to the po
    file. Set outfile to write updated po file to an another file. Set
    to `-\' for writing to standard out. If docstrings is False
    docstrings flag will removed. Set verbose to False to suppress
    progress indicators. <kwds> is used to pass options from the
    command line interface.
    '''
    global option
    option = kwds.get('option', Options(update = update,
                                        outfile = outfile,
                                        docstrings = docstrings,
                                        verbose = verbose))
    def_msgs = parse(def_file, 'msgstr')
    ref_msgs = parse(ref_file, 'msgid')
    if verbose and not __name__ == '__main__':
        print >> sys.stderr, 'Merging %s with %s' % (ref_file, def_file)
    updated_lines = match(def_msgs, ref_msgs)
    if option.verbose:
        print >> sys.stderr, ' done.'
    if not option.update:
        write(updated_lines, option.outfile)
    elif option.update and changes(updated_lines, io(def_file).readlines()):
        write(updated_lines, def_file)
        
def merge_dir(directory, pot = False, include = [], exclude = [],
              verbose = True):
    '''
    Tries to merge a directory of po files. Uses simple glob to find
    po files and pot file. The parameter <pot> can be used to specify
    the pot file in the directory. If the list <include> is given only
    files in this list is merged. Use the list <exclude> to exclude
    files to be merged. This function is only useful if po files and
    pot file are in the same directory. Set <verbose> to get
    information when running.
    '''
    if directory[-1] == '/':
        directory = os.path.dirname(directory)
    if pot:
        pot = os.path.basename(pot)
    else:
        pot = glob.glob('%s/*.pot' % directory)
        if not pot:
            raise MsgmergeError('No pot file found.')
        elif len(pot) > 1:
            raise MsgmergeError('More than one pot file found: %s.' % pot)
        pot = os.path.basename(pot[0])
    
    if not include:
        pos = glob.glob('%s/*po' % directory)
        if not len(pos) > 1:
            raise MsgmergeError('No po file(s) found.')
        pos = [ os.path.basename(po) for po in pos ]
    else:
        pos = [ os.path.basename(po) for po in include ]
    
    for po in exclude:
        try:
            pos.remove(po)
        except ValueError:
            pass
    format = '%s/%s'
    for po in pos:
        try:
            merge(format % (directory, po), format % (directory, pot),
                  update = True, verbose = verbose,
                  outfile = format % (directory, po))
        except MsgmergeError, err:            
            if verbose:
                print >> sys.stderr, '%s Not updated.' % err
            else:
                print >> sys.stderr, '%s %s not updated.' % (err, po)

if __name__ == '__main__':
    cmdline()



""" FeedItems.py

Straw module for handling items that belongs to a feed. This modules is
responsible for adding,cutting, and deleting items of the current feed.
"""
__copyright__ = "Copyright (c) 2002-2005 Free Software Foundation, Inc."
__license__ = """GNU General Public License

This program is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA. """


import QueueDict
import Config
import Feed
import Event
import ItemStore
import error
import locale

class FeedItems(object):
    def __init__(self, feed):
        self._feed = feed
        self._items = QueueDict.ItemQueue()
        self._init_items()
        config = Config.get_instance()
        config.signal_connect(Event.ItemOrderChangedSignal, self.item_order_changed)
        config.signal_connect(Event.NumberOfItemsStoredChangedSignal, self.items_stored_changed)

    def _init_items(self):
        config = Config.get_instance()
        self._items.clear()
        self._items.sort_order = config.item_order
        self._prefs_items_stored = config.number_of_items_stored
        self._idseq = 0
        self._number_of_items = None
        self._loaded = False


    @apply
    def number_of_unread():
        doc = ""
        def fget(self):
            return self._feed.n_items_unread
        def fset(self, n):
            self._feed.n_items_unread = n
        return property(**locals())

    @apply
    def number_of_items():
        doc = ""
        def fget(self):
            if self._number_of_items:
                return self._number_of_items
            if self._loaded:
                error.log("was loaded but number_of_items was None!")
                return 0
        def fset(self, n):
            self._number_of_items = n
        return property(**locals())

    def item_order_changed(self, event):
        self._items.sort_order = event.sender.item_order
        self._feed.signal_refresh_display()

    def items_stored_changed(self, event):
        self._prefs_items_stored = event.sender.number_of_items_stored
        self._feed.signal_refresh_display()

    def add_items(self, new_items):
        cutpoint = self._get_cutpoint()
        items = sorted(new_items, cmp=self._cmp)
        newitems = []
        removed_items = []
        for x, item in enumerate(items):
            if item in self._items:
                continue
            if not item.id:
                self._idseq += 1
                item.id = self._idseq
            item.feed = self._feed
            if self._number_of_items < cutpoint:
                self._items.append(item)
            else:
                # no point on increasing the number of items here since it's
                # already over the cutpoint
                ritem = self._items.replace(item)
                removed_items.append(ritem)
            newitems.append(item)
            item.signal_connect(Event.ItemReadSignal, self.item_read)
            item.signal_connect(Event.ItemStickySignal, self._feed.forward_signal)
        self._idseq = max(self._items.keys())
        self.number_of_unread = len([i for i in self._items.itervalues() if not i.seen])
        self.number_of_items = len(self._items)

        if removed_items:
            self._feed.signal_deleted_item(removed_items)
        if newitems:
            self._feed.signal_new_items(newitems)
        self._feed.signal_refresh_display()

    def restore_items(self, items):
        """
        Restores the given items in an ordered form

        # should we check for duplicates here? Or should that be done
        # when new items arrive?

        """
        items = self._sort(items)
        no_restore = []
        cutpoint = self._get_cutpoint()
        for x,item in enumerate(items):
            item.feed = self._feed
            if item.sticky or x <= cutpoint:
                item.signal_connect(Event.ItemReadSignal, self.item_read)
                item.signal_connect(Event.ItemStickySignal, self._feed.forward_signal)
                self._items.append(item)
                continue
            item.clean_up()
            no_restore.append(item)
        self._idseq = max(self._items.keys())
        self.number_of_unread = len([i for i in self._items.itervalues() if not i.seen])
        self.number_of_items = len(self._items)
        if no_restore:
            self._feed.signal_deleted_item(no_restore)

    def _disconnect_item_signals(self, item):
        item.signal_disconnect(Event.ItemReadSignal, self.item_read)
        item.signal_disconnect(Event.ItemStickySignal, self._feed.forward_signal)

    def item_read(self, signal):
        if signal.sender.seen:
            change = -1
        else:
            change = 1
        self.number_of_unread = self.number_of_unread + change
        self._feed.forward_signal(signal)

    def get_items(self):
        if not self._loaded:
            self.load()
        return self._items.values()

    def get_item_index(self, item):
        if not self._loaded:
            self.load()
        idx = self._items.index(item.id)
        return idx

    def mark_all_read(self):
        if not self._loaded:
            self.load()
        keys = self._items.keys()
        changed = [(keys.index(key), value) for key, value in self._items if value.set_seen_quiet()]
        self.number_of_unread = 0
        self._feed.signal_all_items_read(changed)
        return

    def load(self):
        """
        Loads and restores the items from the data store

        """
        if self._loaded:
            return False
        itemstore = ItemStore.get_instance()
        items = itemstore.read_feed_items(self._feed)
        if items:
            #error.log("load ", self._feed.title, ", number of items: ", len(items), ", number of unread before restore: ", self.number_of_unread)
            self.restore_items(items)
            self._loaded = True
        else:
            #print "No items, not loading ", self._feed.title
            self._loaded = False
        return self._loaded

    def unload(self):
        """
        Unloads the items by disconnecting the signals and reinitialising the
        instance variables

        TODO: unload seems to lose some circular references. garbage collector
        will find them, though, so maybe it's not a problem.
        """
        if not self._loaded:
            return
        for key, item in self._items:
            self._disconnect_item_signals(item)
        self._init_items()

    def _get_cutpoint(self):
        """
        Returns the current cutpoint

        """
        if self._feed.number_of_items_stored == Feed.Feed.DEFAULT:
            cutpoint = self._prefs_items_stored
        else:
            cutpoint = self._feed.number_of_items_stored
        return cutpoint

    def _sort(self, items):
        """
        Sorts the given items according to the sort order
        """
        items.sort(self._cmp)
        if self._items.sort_order:
            items.reverse()
        return items

    def _cmp(self, a, b):
        """
        Comparator method to compare items based on the item's pub_date attribute

        If an item doesn't have a pub_date, it uses title and prioritizes the
        unread items
        """
        try:
            return cmp(a.pub_date, b.pub_date)
        except AttributeError:
            return locale.strcoll(a.title, b.title) and not a.seen or not b.seen



""" Tray.py

Module that puts Straw in the tray notification area.
"""
__copyright__ = "Copyright (c) 2004-2005 Free Software Foundation, Inc."
__author__ = "Olivier Crete <tester@tester.ca>"
__license__ = """ GNU General Public License

This program is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA. """


import os
import os.path
from xml.sax import saxutils
import pygtk
pygtk.require('2.0')
import gtk
import FeedList
import Event
import utils
import constants
import subscribe
import egg
import egg.trayicon

class Tray:
    def __init__(self, ):
        self.tray = egg.trayicon.TrayIcon(constants.APPNAME);
        self._eventbox = gtk.EventBox()
        self.tray.add(self._eventbox)
        self.connect("drag-data-received", self._on_drag_data_received)
        self._eventbox.drag_dest_set(
            gtk.DEST_DEFAULT_ALL,
            [('_NETSCAPE_URL', 0, 0),('text/uri-list ', 0, 1),('x-url/http', 0, 2)],
            gtk.gdk.ACTION_COPY | gtk.gdk.ACTION_MOVE)
        image = gtk.Image()
        iconfile = os.path.normpath(utils.find_image_dir() + "/straw.png")
        pixbuf = gtk.gdk.pixbuf_new_from_file(iconfile)
        scaled_buf = pixbuf.scale_simple(16,16,gtk.gdk.INTERP_BILINEAR)
        image.set_from_pixbuf(scaled_buf)
        self._eventbox.add(image)

        self._tooltip = gtk.Tooltips()
        self.unread_count = 0
        feedlist = FeedList.get_instance()
        feedlist.signal_connect(Event.AllItemsReadSignal,
                                lambda signal: self._update(feedlist))
        feedlist.signal_connect(Event.ItemReadSignal,
                                      lambda signal: self._update(feedlist))
        feedlist.signal_connect(Event.ItemsAddedSignal,
                                      lambda signal: self._update(feedlist))
        feedlist.signal_connect(Event.FeedsChangedSignal,
                                      lambda signal: self._update(feedlist))
        self.tray.show_all()

    def _update(self,flist):
        uritems = urfeeds = 0
        for ur in [f.number_of_unread for f in flist.flatten_list()]:
            if ur:
                uritems += ur
                urfeeds += 1
        if uritems == self.unread_count:
            return
        self.unread_count = uritems
        if uritems:
            self._tooltip.set_tip(self.tray, _("%d new items")%uritems)
            self.tray.show_all()
        else:
            self.tray.hide()
        return

    def connect(self, event, handler):
        self._eventbox.connect(event, handler)
        return

    def _on_drag_data_received(self, widget, context, x, y, data, info, timestamp):
        if data and data.format == 8:
            url = data.data.split("\n")[0]
            subscribe.show(url="%s" % url)
            context.finish(True, False, timestamp)
        else:
            context.finish(False, False, timestamp)
        return



""" ItemList.py

Handles listing of feed items in a view (i.e. GTK+ TreeView)
"""
__copyright__ = "Copyright (c) 2002-2005 Free Software Foundation, Inc."
__license__ = """GNU General Public License

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.
"""

from xml.sax import saxutils
import pygtk
pygtk.require('2.0')
import gobject
import gtk
import pango
import Event
import MVP
import error
import utils

class Column:
    """
    Constants for the item list treeview columns
    """
    TITLE = 0
    STICKY = 1
    ITEM = 2
    WEIGHT = 3
    STICKY_FLAG = 4
    FOREGROUND = 5

class ItemsView(MVP.WidgetView):
    ui = """
    <ui>
        <popup name=\"itemlist_popup\">
            <menuitem action=\"mark_as_unread\"/>
        </popup>
    </ui>
    """

    def _initialize(self):
        self._widget.connect("button_press_event",self._on_button_press_event)
        self._widget.connect("popup-menu", self._on_popup_menu)
        self._widget.connect("row-activated", self._on_row_activated)
        self._popup = None

    def scroll_to_cell(self, path):
        selection = self._widget.get_selection()
        selection.select_path(path)
        self._widget.scroll_to_cell(path)
        self._widget.set_cursor(path)
        return

    def get_widget_selection(self):
        return self._widget.get_selection()

    def _create_popup(self):
        actions = [
            ("mark_as_unread", gtk.STOCK_CONVERT, "Mark as _Unread", None,
             "Mark this item as unread", self._on_menu_mark_as_unread_activate)
            ]
        self._uimanager = gtk.UIManager()
        actiongroup = gtk.ActionGroup("ItemListActions")
        actiongroup.add_actions(actions)
        self._uimanager.insert_action_group(actiongroup, 0)
        self._uimanager.add_ui_from_string(self.ui)
        self._popup = self._uimanager.get_widget('/itemlist_popup')

    def _on_popup_menu(self, treeview, *args):
        self._popup.popup(None,None,None,0,0)

    def _on_button_press_event(self, treeview, event):
        val = 0
        if event.button == 3:
            x = int(event.x)
            y = int(event.y)
            time = gtk.get_current_event_time()
            path = treeview.get_path_at_pos(x, y)
            if path is None:
                return 1
            path, col, cellx, celly = path
            treeview.grab_focus()
            treeview.set_cursor( path, col, 0)
            self._popup.popup(None, None, None, event.button, time)
            val = 1
        return val

    def _on_menu_mark_as_unread_activate(self, *args):
        self._presenter.mark_selected_item_as_unread()

    def _item_selection_changed(self, selection):
        selected = selection.get_selected()
        if selected:
            self._presenter.selection_changed(selected)
        return

    def _on_row_activated(self, treeview, path, column):
        model = self._widget.get_model()
        try: treeiter = model.get_iter(path)
        except ValueError:
            return
        link = model.get_value(treeiter, Column.ITEM).link
        if link:
            utils.url_show(link)
        return

class ItemListView(ItemsView):
    """
    Widget: gtk.Treeview
    Model: ListStore
    Presenter: ItemListPresenter
    """
    def _initialize(self):
        ItemsView._initialize(self)
        self._widget.set_rules_hint(False)
        self._create_popup()
        selection = self._widget.get_selection()
        selection.connect("changed", self._item_selection_changed)

    def _model_set(self):
        self._widget.set_model(self._model)
        self._create_columns()

    def _sticky_toggled(self, cell, path):
        self._presenter.sticky_toggled(cell, path,
                                       Column.ITEM, Column.STICKY_FLAG)
        return

    def _create_columns(self):
        renderer = gtk.CellRendererToggle()
        column = gtk.TreeViewColumn(_('Keep'), renderer,
                                    active=Column.STICKY_FLAG)
        column.set_resizable(True)
        column.set_reorderable(True)
        self._widget.append_column(column)
        renderer.connect('toggled', self._sticky_toggled)

        renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn(_('_Title'), renderer,
                                    text=Column.TITLE,
                                    foreground=Column.FOREGROUND,
                                    weight=Column.WEIGHT)
        column.set_resizable(True)
        column.set_reorderable(True)
        self._widget.append_column(column)
        return

class ItemsPresenter(MVP.BasicPresenter):
    def _connect(self):
        self.initialize_slots(Event.ItemSelectionChangedSignal)
        self._selected_item = None
        return

    def selection_changed(self, selected):
        model, treeiter = selected
        if not treeiter: return
        assert model is self._model
        path = model.get_path(treeiter)
        self._item_selected(path)
        self._view.scroll_to_cell(path)

    def _item_selected(self, path):
        item = self._model[path][Column.ITEM]
        if self._selected_item is not item:
            self._selected_item = item
            if not self._selected_item.seen:
                self._selected_item.seen = 1
            self.emit_signal(Event.ItemSelectionChangedSignal(self,self._selected_item))
        return

    def mark_selected_item_as_unread(self):
        self._selected_item.seen = 0
        self._mark_item(self._selected_item)

    def display_empty_feed(self):
        self._model.clear()
        return

    def _set_column_weight(self, item, path):
        tree_iter = self._model.get_iter(path)
        weight = (pango.WEIGHT_BOLD, pango.WEIGHT_NORMAL)[item.seen]
        colour = ("#0000FF", "#000000")[item.seen]
        self._model.set_value(tree_iter, Column.FOREGROUND, colour)
        self._model.set_value(tree_iter, Column.WEIGHT, weight)
        return

    def _mark_item(self, item, index=None):
        """
        Marks an item as 'READ'

        Subclass SHOULD implement this
        """
        raise NotImplemented

class ItemListPresenter(ItemsPresenter):
    def _initialize(self):
        model = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_OBJECT,
                              gobject.TYPE_PYOBJECT, gobject.TYPE_INT,
                              gobject.TYPE_BOOLEAN, gobject.TYPE_STRING)
        self.model = model
        self._connect()
        self._selected_feed = None
        selection = self._view.get_widget_selection()
        selection.set_mode(gtk.SELECTION_SINGLE)

    def select_previous_item(self):
        """
        Selects the item before the current selection. If there
        is no current selection, selects the last item. If there is no
        previous row, returns False.
        """
        selection = self._view.get_widget_selection()
        model, treeiter = selection.get_selected()
        # select the first item if there is no selection
        if not treeiter:
            treeiter = model.get_iter_first()
        path = self._model.get_path(treeiter)
        # check if there's an item
        if not path:
            return False
        prev_path = path[-1]-1
        # we don't cycle through the items so return False if there are no
        # more items to go to
        if prev_path < 0:
            return False
        selection.select_path(path[-1]-1,)
        return True

    def select_next_item(self):
        """
        Selects the item after the current selection. If there is no current
        selection, selects the first item. If there is no next row, returns False.
        """
        selection = self._view.get_widget_selection()
        model, treeiter = selection.get_selected()
        if not treeiter:
            treeiter = model.get_iter_first()
        next_iter = model.iter_next(treeiter)
        if not next_iter or not self._model.iter_is_valid(treeiter):
            return False
        next_path = model.get_path(next_iter)
        selection.select_path(next_path)
        return True

    def select_last_item(self):
        """
        Selects the last item in this list.
        """
        selection = self._view.get_widget_selection()
        selection.select_path(len(self._model) - 1)
        return True

    def select_next_unread_item(self):
        """
        Selects the first unread item after the current selection. If there is
        no current selection, or if there are no unread items, returns False..

        By returning False, the caller can either go to the next feed, or go
        to the next feed with an unread item.
        """
        has_unread = False
        selection = self._view.get_widget_selection()
        selected_row = selection.get_selected()
        # check if we have a selection. if none,
        # start searching from the first item
        if not selected_row[1]:
            treerow = self._model[0]
        else:
            model, treeiter = selected_row
            if not treeiter:
                return has_unread
            nextiter = model.iter_next(treeiter)
            if not nextiter: # no more rows to iterate
                return has_unread
            treerow = self._model[model.get_path(nextiter)]
        while(treerow):
            item = treerow[Column.ITEM]
            if not item.seen:
                has_unread = True
                selection.select_path(treerow.path)
                break
            treerow = treerow.next
        return has_unread


    def sticky_toggled(self, cell, path, item_column, stickyflag_column):
        treeiter = self._model.get_iter((int(path),))
        item = self._model.get_value(treeiter, item_column)
        item.sticky = not item.sticky
        self._model.set(treeiter, stickyflag_column, item.sticky)

    def display_feed_items(self, feed, select_first=1):
        redisplay = self._selected_feed is feed
        if self._selected_feed and not redisplay:
            self._selected_feed.signal_disconnect(Event.RefreshFeedDisplaySignal,
                                         self._feed_order_changed)
            self._selected_feed.signal_disconnect(Event.AllItemsReadSignal,
                                         self._all_items_read)
            self._selected_feed.signal_disconnect(Event.ItemReadSignal,
                                         self._item_read)
        self._selected_feed = feed
        if not redisplay:
            self._selected_feed.signal_connect(Event.RefreshFeedDisplaySignal,
                                self._feed_order_changed)
            self._selected_feed.signal_connect(Event.AllItemsReadSignal,
                                self._all_items_read)
            self._selected_feed.signal_connect(Event.ItemReadSignal,
                                self._item_read)
        count = self._render_feed(self._selected_feed)
        if not redisplay and count:
            if not self.select_next_unread_item():
                selection = self._view.get_widget_selection()
                selection.select_path((0,))
        return

    def _feed_order_changed(self, event):
        if event.sender is self._selected_feed:
            item = None
            selection = self._view.get_widget_selection()
            model, treeiter = selection.get_selected()
            if treeiter is not None:
                path = model.get_path(treeiter)
                item = model[path[0]][Column.ITEM]
            self._render_feed(event.sender)
            if item:
                path = (item.feed.get_item_index(item),)
            else:
                path = (0,) # first item
            selection.select_path(path)

    def _all_items_read(self, signal):
        for index, item in signal.changed:
            self._mark_item(item, index)

    def _item_read(self, signal):
        self._mark_item(signal.item)

    def _render_feed(self, feed):
        self._model.clear()
        curr_iter = None
        for item in feed.items:
            weight = (pango.WEIGHT_BOLD, pango.WEIGHT_NORMAL)[item.seen]
            colour = ("#0000FF", "#000000")[item.seen]
            treeiter = self._model.append()
            self._model.set(treeiter,
                            Column.TITLE, item.title,
                            Column.ITEM, item,
                            Column.WEIGHT, weight,
                            Column.STICKY_FLAG, item.sticky,
                            Column.FOREGROUND, colour)
            if item is self._selected_item:
                curr_iter = treeiter

        if curr_iter:
            path = self._model.get_path(curr_iter)
            self._view.scroll_to_cell(path)
        return len(feed.items)

    def _mark_item(self, item, path=None):
        if path is None:
            path = (item.feed.get_item_index(item),)
        self._set_column_weight(item, path)
        return



""" subscribe.py

Module for handling feed subscription process.
"""
__copyright__ = "Copyright (c) 2002-2005 Free Software Foundation, Inc."
__license__ = """ GNU General Public License

This program is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA. """


import urlparse
import urllib
from xml.sax import saxutils
import pygtk
pygtk.require('2.0')
import gobject
import gtk
import gtk.glade
import gettext
from Feed import Feed
import FeedList
import feedfinder
import error
import Event
import utils
import SummaryParser
import ParsedSummary
import dialogs
import time
import URLFetch
import NetworkConstants
import MVP
import Config

class FeedLocationView(MVP.GladeView):
    def _initialize(self):
        self._window = self._widget.get_widget("subscribe_window")
        self._notebook = self._widget.get_widget("subscribe_notebook")
        self._entry_url = self._widget.get_widget("subscribe_location_entry")
        self._entry_url.grab_focus()

    def _on_subscribe_location_entry_key_press_event(self, widget, event):
        if event.keyval == gtk.keysyms.Return:
            self.find_feed(self._entry_url.get_text())

    def _on_main_ok_clicked(self, *args):
        self.find_feed(self._entry_url.get_text())

    def _on_main_cancel_clicked(self, *args):
        self._window.hide()

    def find_feed(self, url):
        config = Config.get_instance()
        if not url:
            self._window.hide()
            dialogs.report_error(_("Feed URL Not Provided"),
                                 _("Please provide the URL of the feed you are trying to subscribe to."),
                                 parent=self._window)
            self._window.present()
            self._notebook.set_current_page(Page.LOCATION)
            return
        elif config.offline:
            self._window.hide()
            response = dialogs.report_offline_status(self._window)
            if response == gtk.RESPONSE_CANCEL:
                self._notebook.set_current_page(Page.LOCATION)
                return
            config.offline = not config.offline
        self._window.present()
        try:
            url, uname, pword = self._presenter.split_url(url)
        except TypeError: # bad feed url
            dialogs.report_error(_("Bad Feed URL"),
                                 _("Straw wasn't able to parse '%s'")%url,
                                parent=self._window)
            return
            
        self._presenter.find_feed(url, uname, pword)

    def report_error(self, scheme):
        dialogs.report_error(_("Unsupported Scheme"),
                             _("Subscribing to '%s://' is not supported") % scheme,
                             parent=self._window)

    def clear_entry(self):
        self._entry_url.delete_text(0, -1)

    def set_entry_url(self, url):
        self._entry_url.set_text(url)

class FeedLocationPresenter(MVP.BasicPresenter):
    def _initialize(self):
        self._domain = ""
        self._scheme = None

    def _fix_url(self, url):
        u = urlparse.urlsplit(url.strip())
        # we check if 'scheme' is not empty here because URIs like
        # "google.com" IS valid but, in this case, 'scheme' is empty because
        # urlsplit() expects urls are in the format of scheme://netloc/...
        if not u[0] or (u[0] != "http" and u[0] != "feed"):
            return None
        if u[0] == 'feed':
            u = urlparse.urlsplit(u[2])
        # .. if that happens then we munge the url by adding a // and default
        # to 'http' as the scheme.
        if u[1] == '':
             u = urlparse.urlsplit("//" + url, 'http')
        return u

    def split_url(self, url):
        username, password = "", ""
        scheme, loc, path, query, fragment = self._fix_url(url)

        auth_tokens = loc.split('@')
        if len(auth_tokens) > 1:
            username, password = auth_tokens[0].split(':')
            loc = auth_tokens[1]
        self._domain = loc
        self._scheme = scheme
        url = urlparse.urlunsplit((scheme, loc, path, query, fragment))
        return (url, username, password)

    def find_feed(self, url, username="", password=""):
        if self._scheme and self._scheme != 'http':
            self._view.report_error(self._scheme)
        self._model.poll(url, username, password)

    def get_feed_domain(self):
        return self._domain

    def clear_entry(self):
        self.view.clear_entry()

    def set_location(self):
        def _clipboard_cb(cboard, text, data=None):
            if text:
                b = self._fix_url(text)
                if b and b[0] == "http":
                    u = urlparse.urlunsplit(b)
                    self.view.set_entry_url(u)
        clipboard = gtk.clipboard_get(selection="CLIPBOARD")
        clipboard.request_text(_clipboard_cb, None)


class AuthView(MVP.GladeView):
    def __init__(self, widget, url):
        MVP.GladeView.__init__(self, widget)
        self._url = url

    def _initialize(self):
        self._username_entry = self._widget.get_widget("subscribe_username_entry")
        self._password_entry = self._widget.get_widget("subscribe_password_entry")
        self._username_entry.grab_focus()

    def _on_auth_ok_clicked(self, *args):
        username = self._username_entry.get_text()
        password = self._password_entry.get_text()
        self._presenter.find_feed(self._url, username, password)

class AuthPresenter(FeedLocationPresenter):
    def find_feed(self, url, username, password):
        self._model.poll(url, username, password)

class FeedsTreeView(MVP.GladeView):
    def _initialize(self):
        self._treeview = self._widget.get_widget("feeds_treeview")
        self._toggle_renderer = gtk.CellRendererToggle()
        column = gtk.TreeViewColumn(_('Subscribe'), self._toggle_renderer,
                                    active=FeedColumn.STATUS_FLAG)
        self._treeview.append_column(column)

        renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn(_('Title'), renderer,
                                    markup=FeedColumn.TITLE)
        self._treeview.append_column(column)

    def get_toggle_renderer(self):
        return self._toggle_renderer

    def get_widget(self, name):
        return self._widget.get_widget(name)

    def _model_set(self):
        self._treeview.set_model(self._model)

class FeedsPresenter(MVP.BasicPresenter):
    def _initialize(self):
        self._channel_parser_func = SummaryParser.parse_channel_info
        self._ok_button = self._view.get_widget("mf_ok")
        self._ok_button.set_sensitive(False)
        toggle_renderer = self._view.get_toggle_renderer()
        toggle_renderer.connect("toggled", self._feed_toggled)

    def _feed_toggled(self, cell, path):
        self._model.toggle_feed(cell, path)
        if self._model.get_selected_feeds_count():
            self._ok_button.set_sensitive(True)
        else:
            self._ok_button.set_sensitive(False)

    def set_parser_func(self, func):
        self._channel_parser_func = func

    def init_model(self, feeds):
        self._model.clear()
        for url, data in feeds.items():
            ps = ParsedSummary.ParsedSummary()
            parsed_content, encoding = SummaryParser.feedparser_parse(data)
            if parsed_content.version == "": # feed is neither rss nor atom
                continue
            ps = self._channel_parser_func(ps, parsed_content, encoding)
            self._model.add_feed(url, ps)


class FeedsModel(gtk.ListStore):
    def __init__(self):
        gtk.ListStore.__init__(self, gobject.TYPE_OBJECT, gobject.TYPE_STRING,
                               gobject.TYPE_PYOBJECT,
                               gobject.TYPE_BOOLEAN, gobject.TYPE_STRING)
        self._selected_feeds = list()

    def add_feed(self, url, parsed):
        info = "<b>%s</b>\n%s" % (saxutils.escape(parsed.title),
                                  saxutils.escape(parsed.description))
        title = "%s\n%s" % (info.strip(),
                            saxutils.escape(url))
        iterator = self.append()
        self.set(iterator, FeedColumn.TITLE, title,
                 FeedColumn.PARSEDSUMMARY, parsed,
                 FeedColumn.STATUS_FLAG, False, # Items not ticked by default
                 FeedColumn.URL, url)

    def toggle_feed(self, cell, path):
        iterator = self.get_iter((int(path),))
        url = self.get_value(iterator, FeedColumn.URL)
        sticky_value = not self.get_value(iterator, FeedColumn.STATUS_FLAG)
        if sticky_value:
            self._selected_feeds.append(url)
        else:
            self._selected_feeds.remove(url)
        self.set(iterator, FeedColumn.STATUS_FLAG, sticky_value)

    def get_selected_feeds(self):
        feeds = self._selected_feeds[:]
        del self._selected_feeds[:]
        return feeds

    def get_selected_feeds_count(self):
        return len(self._selected_feeds)

class Subscribe:
    def __init__(self, xml, parent):
        self._widget = xml
        self._stoppers = list()
        self._poller = Poller()
        self._poller.signal_connect(PollStartSignal, self.poll_start_cb)
        self._poller.signal_connect(AuthNeededSignal, self.auth_needed_cb)
        self._poller.signal_connect(PollDoneSignal, self.poll_done_cb)
        self._poller.signal_connect(PollFailedSignal, self.poll_failed_cb)
        self._feed_location_presenter = FeedLocationPresenter(
            model= self._poller, view = FeedLocationView(xml))
        self._feeds_selection_presenter = FeedsPresenter(
            model = FeedsModel(), view = FeedsTreeView(self._widget))
        self._window = self._widget.get_widget("subscribe_window")
        self._window.set_transient_for(parent)
        self._window.connect('delete-event', self.on_subscribe_window_delete_event)
        self._notebook = self._widget.get_widget("subscribe_notebook")
        self._progress_bar = self._widget.get_widget("subscribe_progress_bar")
        self._progress_label = self._widget.get_widget("subscribe_progress_label")
        self._widget.get_widget("progress_cancel").connect("clicked",
                                                           self.close_window)
        self._widget.get_widget("main_cancel").connect("clicked",
                                                       self.close_window)
        self._widget.get_widget("auth_cancel").connect("clicked",
                                                       self.close_window)
        self._widget.get_widget("mf_cancel").connect("clicked",
                                                       self.close_window)
        self._widget.get_widget("mf_ok").connect("clicked", self.feed_selection_done)
        self._feeds = dict()
        self._username = ""
        self._password = ""
        self._update_progress = False

    def _get_poller(self):
        return self._poller

    poller = property(_get_poller)

    def show(self):
        self._notebook.set_current_page(Page.LOCATION)
        self._feed_location_presenter.set_location()
        self._window.show()

    def hide(self):
        self._window.hide()

    def on_subscribe_window_delete_event(self, *args):
        self.close_window()
        return gtk.TRUE

    def set_url_dnd(self, url):
        self._window.show()
        self._feed_location_presenter.view.find_feed(url)

    def _reset(self):
        self._notebook.set_current_page(Page.LOCATION)
        self._feeds.clear()
        self._feed_location_presenter.clear_entry()
        self._update_progress = False

    def close_window(self, *args):
        self._reset()
        self.remove_stoppers()
        self.hide()

    def remove_stoppers(self):
        [stopper() for stopper in self._stoppers if stopper is not None]
        del self._stoppers[:]

    def poll_start_cb(self, event):
        self._stoppers.append(event.stopper)
        self._username, self._password = event.username, event.password
        domain = self._feed_location_presenter.get_feed_domain()
        if self._notebook.get_current_page() != Page.PROGRESS:
            self._window.set_title(_("Searching %s" % domain))
            self._notebook.set_current_page(Page.PROGRESS)
        if not self._update_progress:
            self._start_progress_bar()
            self.progress_set_text(_("Searching %s. This may take a while...") % domain)

    def _start_progress_bar(self):
        self._update_progress = True
        self._progress_bar.set_fraction(0)
        def _update_progress_bar():
            self._progress_bar.pulse()
            if self._update_progress:
                return True
            else: return False
        update = gobject.timeout_add(150, _update_progress_bar)

    def progress_set_text(self, text):
        self._progress_label.set_text("%s" % text)

    def auth_needed_cb(self, event):
        self._notebook.set_current_page(Page.AUTH)
        AuthPresenter(view = AuthView(self._widget, event.url),
                      model = self._poller)

    def poll_done_cb(self, event):
        self.remove_stoppers()
        try:
            feeds, data, data_is_feed = feed_data_check(event.url, event.data)
        except IOError, ioe:
            num_feeds = 0;
            self._display_dialog(_("Unexpected Error Occurred"), str(ioe))
        else:
            num_feeds = len(feeds)
            if num_feeds == 1:
                self._process_feed(feeds, data, data_is_feed)
            elif num_feeds > 1:
                self._process_multi_feed(feeds)
            else:
                self._display_dialog(_("Unable to Find Feeds"),
                                     (_("Straw was unable to find feeds in %s") % event.url))

    def _process_feed(self, feeds, data, data_is_feed):
        feed_url = feeds.pop()
        if data_is_feed:
            self.add_to_subscription(feed_url, data)
            self.close_window()
        else:
            self._poller.poll(feed_url, self._username, self._password)
        return

    def _process_multi_feed(self, feeds):
        num_feeds = len(feeds)
        for url in feeds:
            self._feeds[url] = None
        self._poller.signal_disconnect(PollDoneSignal, self.poll_done_cb)
        unpolled = 1
        def mf_cb(event):
            if not feeds:
                return
            self._feeds[feeds.pop()] = event.data
            unpolled = len([v for v in self._feeds.itervalues() if v is None])
            self.progress_set_text(_("Processing %d of %d feeds") % (num_feeds-unpolled,num_feeds))
            if unpolled == 0:
                self.remove_stoppers()
                self._prepare_feed_selection_page()
        self._poller.signal_connect(PollDoneSignal, mf_cb)
        for feed in feeds:
            self._poller.poll(feed, self._username, self._password)
        return

    def _prepare_feed_selection_page(self):
        self._notebook.set_current_page(Page.FEEDS)
        self._feeds_selection_presenter.init_model(self._feeds)

    def poll_failed_cb(self, event):
        text = ""
        self._update_progress = False
        if event.exception is not None:
            text = str(event.exception)
        elif event.message is not None:
            text = event.message
        self._display_dialog(_("Error While Subscribing"), text)

    def _display_dialog(self, title, text):
        self._window.hide()
        response = dialogs.report_error(_("%s" % title),
                                        _("%s" % text),
                                        parent=self._window)
        self._window.show()
        if response:
            self.remove_stoppers()
            self._notebook.set_current_page(Page.LOCATION)

    def feed_selection_done(self, event):
        sf = self._feeds_selection_presenter.model.get_selected_feeds()
        for url in sf:
            if url in self._feeds:
                self.add_to_subscription(url, self._feeds[url])
        self.close_window()

    def add_to_subscription(self, url, data):
        feed = Feed.create_new_feed("", url, self._username, self._password)
        parsed = SummaryParser.parse(data, feed)
        if parsed.title == "":
            # store url loc in title in case it's empty
            parsed.title = urlparse.urlsplit(url.strip())[1]
        feed.title = utils.convert_entities(parsed.title)
        feed.channel_description = utils.convert_entities(parsed.description)
        feed.access_info = (url, self._username, self._password)
        feed.last_poll = int(time.time())
        category = None # Add to 'All' category
        FeedList.get_instance().append(category, feed)
        feed.router.route_all(None, parsed)
        feed.poll_done()

class PollStartSignal(Event.BaseSignal):
    def __init__(self, sender, stopper, url, username, password):
        Event.BaseSignal.__init__(self,sender)
        self.stopper = stopper
        self.url = url
        self.username = username
        self.password = password

class PollDoneSignal(Event.BaseSignal):
    def __init__(self, sender, url, data):
        Event.BaseSignal.__init__(self,sender)
        self.data = data
        self.url = url

class AuthNeededSignal(Event.BaseSignal):
    def __init__(self, sender, url):
        Event.BaseSignal.__init__(self, sender)
        self.url = url

class PollFailedSignal(Event.BaseSignal):
    def __init__(self, sender, url, exception=None, message=None):
        Event.BaseSignal.__init__(self, sender)
        self.url = url
        self.exception = exception
        self.message = message

class Poller(Event.SignalEmitter):
    def __init__(self):
        Event.SignalEmitter.__init__(self)
        self.initialize_slots(PollStartSignal,
                              AuthNeededSignal,
                              PollDoneSignal,
                              PollFailedSignal)
        self._request_handler = URLFetch.connection_manager
        self._request_priority = NetworkConstants.PRIORITY_RSS
        self._url = None
        self._username = None
        self._password = None

    def set_request_handler(self, handler):
        # currently for testing purposes
        self._request_handler = handler

    def set_request_priority(self, priority):
        self._request_priority = priority

    def poll(self, url, username="", password=""):
        url = url.strip()
        self._url = feedfinder.makeFullURI(url)
        self._username = username
        self._password = password
        headers = {}
        stopper = None

        try:
            stopper = self._request_handler.request(self._url, self,
                                                    headers, self._username, self._password,
                                                    priority=self._request_priority)
        except Exception, e:
            self.http_failed(e)
        else:
            self.emit_signal(PollStartSignal(self, stopper, self._url, self._username, self._password))

    def http_results(self, status, header, data):
        error = None
        if status is None:
            error = _("No Data")
        else:
            code = status[1]
            if code == 401:
                self.emit_signal(AuthNeededSignal(self, self._url))
                return
            elif code >= 402:
                error = "%s: %s" % (code, status[2].strip())

        if not error:
            self.emit_signal(PollDoneSignal(self, self._url, data))
        else:
            self.http_failed(msg=error)
        return

    def http_failed(self, exception=None, msg=None):
        self.emit_signal(PollFailedSignal(self, self._url, exception, msg))

    def http_permanent_redirect(self, newloc):
        self._url = newloc
        self.poll(self._url, self._username, self._password)

    def operation_stopped(self):
        pass

class Page:
    LOCATION = 0
    PROGRESS = 1
    AUTH = 2
    FEEDS = 3

class FeedColumn:
    SUBSCRIBE = 0
    TITLE = 1
    PARSEDSUMMARY = 2
    STATUS_FLAG = 3
    URL = 4

def feed_data_check(url, data):
    feeds = list()
    data_is_feed = False
    if feedfinder.couldBeFeedData(data):
        feeds.append(url)
        data_is_feed = True
    else:
        feeds = feedfinder.getLinks(data, url)
        feeds = filter(feedfinder.isFeed, feeds)
        if not feeds:
            links = feedfinder.getALinks(data, url)
            locallinks = feedfinder.getLocalLinks(links, url)
            # XXX: feedfinder.isfeed() blocks.
            feeds = filter(feedfinder.isFeed, filter(feedfinder.isFeedLink, locallinks))
            if not feeds:
                feeds = filter(feedfinder.isFeed, filter(feedfinder.isXMLRelatedLink, locallinks))
            if not feeds:
                feeds = filter(feedfinder.isFeed, filter(feedfinder.isFeedLink, links))
            if not feeds:
                feeds = filter(feedfinder.isFeed,
                               filter(feedfinder.isXMLRelatedLink, links))
    return (uniq(feeds), data, data_is_feed)

# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/52560
# Author: Raymond Hettinger
def uniq(alist):    # Fastest without order preserving
    set = {}
    map(set.__setitem__, alist, [])
    return set.keys()

instance = None
def _get_instance(parent=None):
    global instance
    if instance is None:
        glade_file = utils.find_glade_file()
        xml = gtk.glade.XML(glade_file, "subscribe_window",
                            gettext.textdomain())
        instance = Subscribe(xml, parent)
    return instance

def show(parent=None, url=None):
    sd = _get_instance(parent)
    if url is None:
        sd.show()
    else:
        sd.set_url_dnd(url)
